'''
An implementation of the differential evolution algorithm for fitting.
'''
import threading
import multiprocessing as processing
import pickle
import random as random_mod
import time
from dataclasses import dataclass
from logging import debug
from numpy import *

from .core.config import BaseConfig
from .core.custom_logging import iprint
from .core.Simplex import Simplex
from .exceptions import ErrorBarError
from .model import Model
from .solver_basis import GenxOptimizer, GenxOptimizerCallback, SolverParameterInfo, SolverResultInfo, SolverUpdateInfo


__mpi_loaded__=False
__parallel_loaded__=True
_cpu_count=processing.cpu_count()

try:
    from mpi4py import MPI as mpi
except ImportError:
    rank=0
    __mpi_loaded__=False
    mpi=None
else:
    __mpi_loaded__=True
    comm=mpi.COMM_WORLD
    size=comm.Get_size()
    rank=comm.Get_rank()


class DiffEvDefaultCallbacks(GenxOptimizerCallback):

    def text_output(self, text):
        iprint(text, flush=True)

    def plot_output(self, update_data):
        pass

    def parameter_output(self, param_info):
        pass

    def fitting_ended(self, result_data):
        pass

    def autosave(self):
        pass


@dataclass
class DiffEvConfig(BaseConfig):
    section='solver'
    km:float=BaseConfig.GParam(0.7, pmin=0.0, pmax=1.0)
    kr:float=BaseConfig.GParam(0.7, pmin=0.0, pmax=1.0)
    allowed_fom_discrepancy:float= 1e-10

    use_pop_mult:bool=True
    pop_size:int=BaseConfig.GParam(50, pmin=5, pmax=10000, label='Fixed size')
    pop_mult:int=BaseConfig.GParam(3, pmin=1, pmax=100, label='Relative size')
    create_trial:str=BaseConfig.GChoice('best_1_bin', ['best_1_bin', 'rand_1_bin', 'best_either_or',
                                                       'rand_either_or', 'jade_best', 'simplex_best_1_bin'],
                                        label='Method')

    use_max_generations:bool=False
    max_generations:int=BaseConfig.GParam(500, pmin=10, pmax=10000, label='Fixed size')
    max_generation_mult:int=BaseConfig.GParam(6, pmin=1, pmax=100, label='Relative size')

    use_start_guess:bool=True
    use_boundaries:bool=True

    max_log_elements:int=BaseConfig.GParam(100000, pmin=1000, pmax=1000000, label=', # elements')
    use_mpi:bool=False
    use_parallel_processing:bool=False
    parallel_processes:int=BaseConfig.GParam(_cpu_count, pmin=2, pmax=_cpu_count, label='# processes')
    parallel_chunksize:int=BaseConfig.GParam(10, pmin=1, pmax=1000, label='items/chunk')

    use_autosave:bool=False
    autosave_interval:int=BaseConfig.GParam(10, pmin=1, pmax=1000, label=', interval')

    save_all_evals:bool=False
    errorbar_level:float=BaseConfig.GParam(1.05, pmin=1.001, pmax=2.0)

    groups={ # for building config dialogs
        'Fitting': [['use_start_guess', 'use_boundaries'], ['use_autosave', 'autosave_interval'],
                    ['save_all_evals', 'max_log_elements']],
        'Differential Evolution':
            ['km', 'kr', 'create_trial',
             ['Population size:', 'use_pop_mult', 'pop_mult', 'pop_size'],
             ['Max. Generations:', 'use_max_generations', 'max_generations', 'max_generation_mult']
             ],
        'Parallel processing': ['use_parallel_processing', 'parallel_processes', 'parallel_chunksize']
        }

class DiffEv(GenxOptimizer):
    '''
    Contains the implementation of the differential evolution algorithm.
    It also contains thread support which is activated by the start_fit 
    function.
    '''
    opt: DiffEvConfig
    model: Model
    fom_log: ndarray
    start_guess: ndarray

    pf=0.5 # probability for mutation
    c=0.07
    simplex_interval = 5  # Interval of running the simplex opt
    simplex_step = 0.05  # first step as a fraction of pop size
    simplex_n = 0.0  # Number of individuals that will be optimized by simplex
    simplex_rel_epsilon = 1000  # The relative epsilon - convergence criteria
    simplex_max_iter = 100  # THe maximum number of simplex runs

    _callbacks: GenxOptimizerCallback=DiffEvDefaultCallbacks()

    def create_mutation_table(self):
        # Mutation schemes implemented
        self.mutation_schemes=[self.best_1_bin, self.rand_1_bin,
                               self.best_either_or, self.rand_either_or,
                               self.jade_best, self.simplex_best_1_bin]

    def __init__(self):
        GenxOptimizer.__init__(self)
        self.create_mutation_table()

        self.model=Model()

        # Definition for the create_trial function
        self.create_trial=self.best_1_bin
        self.update_pop=self.standard_update_pop
        self.init_new_generation=self.standard_init_new_generation

        # Control flags:
        self.running=False  # true if optimization is running
        self.stop=False  # true if the optimization should stop
        self.setup_ok=False  # True if the optimization have been setup
        self.error=None  # None/string if an error ahs occurred

        # Logging variables
        self.fom_log=array([[0, 0]])[0:0]

        self.par_evals=CircBuffer(self.opt.max_log_elements, buffer=array([[]])[0:0])
        self.fom_evals=CircBuffer(self.opt.max_log_elements)

        self.start_guess=array([])
        self.updated_kr=[]
        self.updated_km=[]


    @property
    def n_fom_evals(self):
        return len(self.fom_evals)

    @property
    def method(self):
        return self.create_trial.__name__

    @method.setter
    def method(self, value):
        names=self.methods
        if value in names:
            self.create_trial=self.mutation_schemes[names.index(value)]
        else:
            raise ValueError("Mutation method has to be in %s"%names)

    @property
    def methods(self):
        return [f.__name__ for f in self.mutation_schemes]

    def write_h5group(self, group):
        """
        Write parameters into hdf5 group
        """
        super().write_h5group(group)

        if not self.opt.save_all_evals:
            group['par_evals']=array([])
            group['fom_evals']=array([])
        else:
            group['par_evals']=self.par_evals.array()
            group['fom_evals']=self.fom_evals.array()

    def read_h5group(self, group):
        """
        Read parameters from a hdf5 group
        """
        self.setup_ok=False
        super().read_h5group(group)

        self.par_evals.copy_from(group['par_evals'][()])
        self.fom_evals.copy_from(group['fom_evals'][()])

    def get_start_guess(self):
        return self.start_guess

    def is_running(self):
        return self.running

    def project_evals(self, index):
        return self.par_evals[:, index], self.fom_evals[:]

    def is_fitted(self):
        return len(self.start_guess)>0

    def is_configured(self):
        return self.setup_ok

    def safe_copy(self, other: 'DiffEv'):
        '''
        Does a safe copy of other to this object. Makes copies of everything
        if necessary. The two objects become decoupled.
        '''
        self.opt=other.opt.copy()

        # True if the optimization have been setup
        self.setup_ok=other.setup_ok

        # Logging variables
        self.fom_log=other.fom_log[:]
        self.par_evals.copy_from(other.par_evals)
        self.fom_evals.copy_from(other.fom_evals)

        if self.setup_ok:
            self.n_pop=other.n_pop
            self.max_gen=other.max_gen

            # Starting values setup
            self.pop_vec=other.pop_vec

            self.start_guess=other.start_guess

            self.trial_vec=other.trial_vec
            self.best_vec=other.best_vec

            self.fom_vec=other.fom_vec
            self.best_fom=other.best_fom
            # Not all implementations have these copied within their files
            # Just ignore if an error occur
            try:
                self.n_dim=other.n_dim
                self.par_min=other.par_min
                self.par_max=other.par_max
            except AttributeError:
                debug("Ignoring undefined parameters from DiffEv object in copy due to error:", exc_info=True)

    def pickle_string(self, clear_evals=False):
        '''
        Saves a copy into a pickled string note that the dynamic
        functions will not be saved. For normal use this is taken care of
        outside this class with the config object.
        '''
        cpy=DiffEv()
        cpy.safe_copy(self)
        if clear_evals:
            cpy.par_evals.buffer=cpy.par_evals.buffer[0:0]
            cpy.fom_evals.buffer=cpy.fom_evals.buffer[0:0]
        cpy.create_trial=None
        cpy.update_pop=None
        cpy.init_new_generation=None

        cpy._callbacks=DiffEvDefaultCallbacks()
        cpy.model=None
        cpy.mutation_schemes=None

        return pickle.dumps(cpy)

    def pickle_load(self, pickled_string: bytes):
        '''
        Loads the pickled string into the this object. See pickle_string.
        '''
        obj=pickle.loads(pickled_string, encoding='latin1', errors='ignore')
        obj.create_mutation_table()
        if not hasattr(obj, 'opt'):
            obj.opt=self.opt.copy()
        self.safe_copy(obj)

    def reset(self):
        '''
        Resets the optimizer. Note this has to be run if the optimizer is to
        be restarted.
        '''
        self.setup_ok=False

    def connect_model(self, model_obj):
        '''
        Connects the model [model] to this object. Retrives the function
        that sets the variables  and stores a reference to the model.
        '''
        # Retrieve parameters from the model
        (param_funcs, start_guess, par_min, par_max)=model_obj.get_fit_pars()

        # Control parameter setup
        self.par_min=array(par_min)
        self.par_max=array(par_max)
        self.par_funcs=param_funcs
        self.model=model_obj
        self.n_dim=len(param_funcs)

        if not self.setup_ok:
            self.start_guess=start_guess

    def init_fitting(self, model_obj):
        '''
        Function to run before a new fit is started with start_fit.
        It initilaize the population and sets the limits on the number
        of generation and the population size.
        '''
        self.connect_model(model_obj)
        if self.opt.use_pop_mult:
            self.n_pop=int(self.opt.pop_mult*self.n_dim)
        else:
            self.n_pop=int(self.opt.pop_size)
        if self.opt.use_max_generations:
            self.max_gen=int(self.opt.max_generations)
        else:
            self.max_gen=int(self.opt.max_generation_mult*self.n_dim*self.n_pop)

        # Starting values setup
        self.pop_vec=[self.par_min+random.rand(self.n_dim)*(self.par_max-self.par_min)
                      for _ in range(self.n_pop)]

        if self.opt.use_start_guess:
            self.pop_vec[0]=array(self.start_guess)

        self.trial_vec=[zeros(self.n_dim) for _ in range(self.n_pop)]
        self.best_vec=self.pop_vec[0]

        self.fom_vec=zeros(self.n_dim)
        self.best_fom=1e20

        # Storage area for JADE archives
        self.km_vec=ones(self.n_dim)*self.opt.km
        self.kr_vec=ones(self.n_dim)*self.opt.kr

        # Logging variables
        self.fom_log=array([[0, 1]])[0:0]
        self.par_evals=CircBuffer(self.opt.max_log_elements, buffer=array([self.par_min])[0:0])
        self.fom_evals=CircBuffer(self.opt.max_log_elements)
        # Number of FOM evaluations
        self.n_fom=0

        if rank==0:
            self.text_output('DE initilized')

        # Remember that everything has been setup ok
        self.setup_ok=True

    def init_fom_eval(self):
        '''
        Makes the eval_fom function
        '''
        # Setting up for parallel processing
        if self.opt.use_parallel_processing and __parallel_loaded__:
            self.text_output('Setting up a pool of workers ...')
            self.setup_parallel()
            self.eval_fom=self.calc_trial_fom_parallel
        elif self.opt.use_mpi and __mpi_loaded__:
            self.setup_parallel_mpi()
            self.eval_fom=self.calc_trial_fom_parallel_mpi
        else:
            self.eval_fom=self.calc_trial_fom

    def start_fit(self, model_obj):
        '''
        Starts fitting in a seperate thred.
        '''
        if not self.running:
            # Initialize the parameters to fit
            self.reset()
            self.init_fitting(model_obj)
            self.stop=False
            # Start fitting in a new thread
            threading.Thread(target=self.optimize, daemon=True).start()
            self.running=True
            self.text_output('Starting the fit...')
            return True
        else:
            self.text_output('Fit is already running, stop and then start')
            return False

    def stop_fit(self):
        '''
        Stops the fit if it has been started in a seperate theres 
        by start_fit.
        '''
        if self.running:
            self.stop=True
            self.text_output('Trying to stop the fit...')
        else:
            self.text_output('The fit is not running')

    def resume_fit(self, model_obj):
        '''
        Resumes the fitting if has been stopped with stop_fit.
        '''
        if not self.running:
            self.stop=False
            self.connect_model(model_obj)
            self.init_fom_eval()
            n_dim_old=self.n_dim
            if self.n_dim==n_dim_old:
                threading.Thread(target=self.optimize, daemon=True).start()
                self.text_output('Restarting the fit...')
                self.running=True
                return True
            else:
                self.text_output('The number of parameters has changed'
                                 ' restart the fit.')
                return False
        else:
            self.text_output('Fit is already running, stop and then start')
            return False

    def optimize(self):
        """Method that does the optimization.

        Note that this method does not run in a separate thread.
        For threading use start_fit, stop_fit and resume_fit instead.
        """
        if self.opt.use_mpi:
            self.optimize_mpi()
        else:
            self.optimize_standard()

    def optimize_standard(self):
        '''
        Method implementing the main loop of the differential evolution
        algorithm. Note that this method does not run in a separate thread.
        For threading use start_fit, stop_fit and resume_fit instead.
        '''
        self.running=True
        self.init_fom_eval()

        self.text_output('Calculating start FOM ...')
        self.error=None
        self.n_fom=0

        self.trial_vec=self.pop_vec[:]
        self.eval_fom()
        [self.par_evals.append(vec, axis=0) for vec in self.pop_vec]
        [self.fom_evals.append(vec) for vec in self.trial_fom]
        self.fom_vec=self.trial_fom[:]

        best_index=argmin(self.fom_vec)
        self.best_vec=copy(self.pop_vec[best_index])
        self.best_fom=self.fom_vec[best_index]
        if len(self.fom_log)==0:
            self.fom_log=r_[self.fom_log, \
                            [[len(self.fom_log), self.best_fom]]]
        # Flag to keep track if there has been any improvements
        # in the fit - used for updates
        self.new_best=True

        self.text_output('Going into optimization ...')

        # Update the plot data for any gui or other output
        self.plot_output()
        self.parameter_output()

        # Just making gen live in this scope as well...
        gen=self.fom_log[-1, 0]
        for gen in range(int(self.fom_log[-1, 0])+1, self.max_gen+int(self.fom_log[-1, 0])+1):
            if self.stop:
                break

            t_start=time.time()

            self.init_new_generation(gen)

            # Create the vectors who will be compared to the 
            # population vectors
            [self.create_trial(index) for index in range(self.n_pop)]
            self.eval_fom()
            # Calculate the fom of the trial vectors and update the population
            [self.update_pop(index) for index in range(self.n_pop)]

            # Add the evaluation to the logging
            [self.par_evals.append(vec, axis=0) for vec in self.trial_vec]
            [self.fom_evals.append(vec) for vec in self.trial_fom]

            # Add the best value to the fom log
            self.fom_log=r_[self.fom_log, [[len(self.fom_log), self.best_fom]]]

            # Let the model calculate the simulation of the best.
            sim_fom=self.calc_sim(self.best_vec)

            # Sanity of the model does the simulations fom agree with
            # the best fom
            if abs(sim_fom-self.best_fom)>self.opt.allowed_fom_discrepancy:
                self.text_output('Disagrement between two different fom'
                                 ' evaluations')
                self.error=('The disagreement between two subsequent '
                            'evaluations is larger than %s. Check the '
                            'model for circular assignments.'
                            %self.opt.allowed_fom_discrepancy)
                break

            # Update the plot data for any gui or other output
            self.plot_output()
            self.parameter_output()

            # Time measurement to track the speed
            t=time.time()-t_start
            if t>0:
                speed=self.n_pop/t
            else:
                speed=999999
            self.text_output('FOM: %.3f Generation: %d Speed: %.1f'% \
                             (self.best_fom, gen, speed))

            self.new_best=False
            # Do an autosave if activated and the interval is correct
            if gen%self.opt.autosave_interval==0 and self.opt.use_autosave:
                self.autosave()

        if not self.error:
            self.text_output('Stopped at Generation: %d after %d fom evaluations...'%(gen, gen*self.n_pop))

        # Lets clean up and delete our pool of workers
        if self.opt.use_parallel_processing:
            self.dismount_parallel()
        self.eval_fom=None

        # Now the optimization has stopped
        self.running=False

        # Run application specific clean-up actions
        self.fitting_ended()

    def optimize_mpi(self):
        '''
        Method implementing the main loop of the differential evolution
        algorithm using mpi. This should only be used from the command line.
        The gui can not handle to use mpi.
        '''
        self.running=True
        self.init_fom_eval()

        if rank==0:
            self.text_output('Calculating start FOM ...')
        self.running=True
        self.error=None
        self.n_fom=0
        # Old leftovers before going parallel
        self.fom_vec=[self.calc_fom(vec) for vec in self.pop_vec]
        [self.par_evals.append(vec, axis=0) \
         for vec in self.pop_vec]
        [self.fom_evals.append(vec) for vec in self.fom_vec]

        best_index=argmin(self.fom_vec)
        self.best_vec=copy(self.pop_vec[best_index])
        self.best_fom=self.fom_vec[best_index]
        if len(self.fom_log)==0:
            self.fom_log=r_[self.fom_log, \
                            [[len(self.fom_log), self.best_fom]]]
        # Flag to keep track if there has been any improvements
        # in the fit - used for updates
        self.new_best=True

        if rank==0:
            self.text_output('Going into optimization ...')

        # Update the plot data for any gui or other output
        self.plot_output()
        self.parameter_output()

        # Just making gen live in this scope as well...
        gen=self.fom_log[-1, 0]
        for gen in range(int(self.fom_log[-1, 0])+1, self.max_gen \
                                                     +int(self.fom_log[-1, 0])+1):
            if self.stop:
                break
            t_start=time.time()

            self.init_new_generation(gen)

            # Create the vectors who will be compared to the
            # population vectors
            if rank==0:
                [self.create_trial(index) for index in range(self.n_pop)]
                tmp_trial_vec=self.trial_vec
            else:
                tmp_trial_vec=0
            tmp_trial_vec=comm.bcast(tmp_trial_vec, root=0)
            self.trial_vec=tmp_trial_vec
            self.eval_fom()
            tmp_fom=self.trial_fom
            comm.Barrier()

            # collect forms and reshape them and set the completed tmp_fom to trial_fom
            tmp_fom=comm.gather(tmp_fom, root=0)
            if rank==0:
                tmp_fom_list=[]
                for i in list(tmp_fom):
                    tmp_fom_list=tmp_fom_list+i
                tmp_fom=tmp_fom_list

            tmp_fom=comm.bcast(tmp_fom, root=0)
            self.trial_fom=array(tmp_fom).reshape(self.n_pop, )

            [self.update_pop(index) for index in range(self.n_pop)]

            # Calculate the fom of the trial vectors and update the population
            if rank==0:
                # Add the evaluation to the logging
                [self.par_evals.append(vec, axis=0) for vec in self.trial_vec]
                [self.fom_evals.append(vec) for vec in self.trial_fom]

                # Add the best value to the fom log
                self.fom_log=r_[self.fom_log, [[len(self.fom_log), self.best_fom]]]

                # Let the model calculate the simulation of the best.
                sim_fom=self.calc_sim(self.best_vec)

                # Sanity of the model does the simulations fom agree with
                # the best fom
                if abs(sim_fom-self.best_fom)>self.opt.allowed_fom_discrepancy and rank==0:
                    self.text_output('Disagrement between two different fom'
                                     ' evaluations')
                    self.error=('The disagreement between two subsequent '
                                'evaluations is larger than %s. Check the '
                                'model for circular assignments.'
                                %self.opt.allowed_fom_discrepancy)
                    break

                # Update the plot data for any gui or other output
                self.plot_output()
                self.parameter_output()

                # Let the optimization sleep for a while
                # time.sleep(self.opt.sleep_time)

                # Time measurement to track the speed
                t=time.time()-t_start
                if t>0:
                    speed=self.n_pop/t
                else:
                    speed=999999
                if rank==0:
                    self.text_output('FOM: %.3f Generation: %d Speed: %.1f'%
                                     (self.best_fom, gen, speed))

                self.new_best=False
                # Do an autosave if activated and the interval is correct
                if gen%self.opt.autosave_interval==0 and self.opt.use_autosave:
                    self.autosave()

        if rank==0:
            if not self.error:
                self.text_output('Stopped at Generation: %d after %d fom evaluations...'%(gen, gen*self.n_pop))

        # Lets clean up and delete our pool of workers

        self.eval_fom=None

        # Now the optimization has stopped
        self.running=False

        # Run application specific clean-up actions
        self.fitting_ended()

    def calc_fom(self, vec):
        '''
        Function to calcuate the figure of merit for parameter vector 
        vec.
        '''
        # Set the parameter values
        list(map(lambda func, value: func(value), self.par_funcs, vec))
        fom=self.model.evaluate_fit_func()
        self.n_fom+=1
        return fom

    def calc_trial_fom(self):
        '''
        Function to calculate the fom values for the trial vectors
        '''
        self.trial_fom=[self.calc_fom(vec) for vec in self.trial_vec]

    def calc_sim(self, vec):
        ''' calc_sim(self, vec) --> None
        Function that will evaluate the the data points for
        parameters in vec.
        '''
        # Set the parameter values
        list(map(lambda func, value: func(value), self.par_funcs, vec))

        self.model.evaluate_sim_func()
        return self.model.fom

    def setup_parallel(self):
        '''setup_parallel(self) --> None
        
        setup for parallel proccesing. Creates a pool of workers with
        as many cpus there is available
        '''
        # check if CUDA has been activated
        from .models.lib import paratt, USE_NUMBA
        use_cuda=paratt.Refl.__module__.rsplit('.',1)[1]=='paratt_cuda'
        # reduce numba thread count for numba functions
        if USE_NUMBA:
            numba_procs=max(1, _cpu_count//self.opt.parallel_processes)
        else:
            numba_procs=None
        self.text_output("Starting a pool with %i workers ..."%(self.opt.parallel_processes,))
        self.pool=processing.Pool(processes=self.opt.parallel_processes,
                                  initializer=parallel_init,
                                  initargs=(self.model.pickable_copy(), numba_procs))
        if use_cuda:
            self.pool.apply_async(init_cuda)
        time.sleep(1.0)
        # print "Starting a pool with ", self.opt.parallel_processes, " workers ..."

    def setup_parallel_mpi(self):
        """Inits the number or process used for mpi.
        """
        # check if CUDA has been activated
        from .models.lib import paratt, USE_NUMBA
        use_cuda=paratt.Refl.__module__.rsplit('.',1)[1]=='paratt_cuda'
        # reduce numba thread count for numba functions
        if USE_NUMBA:
            numba_procs=max(1, _cpu_count//self.opt.parallel_processes)
        else:
            numba_procs=None

        if rank==0:
            self.text_output("Inits mpi with %i processes ..."%(size,))
        parallel_init(self.model.pickable_copy(), numba_procs, use_mpi=True)
        time.sleep(0.1)

    def dismount_parallel(self):
        '''
        Used to close the pool and all its processes
        '''
        self.pool.close()
        self.pool.join()

        # del self.pool

    def calc_trial_fom_parallel(self):
        '''
        Function to calculate the fom in parallel using the pool
        '''
        self.trial_fom=self.pool.map(parallel_calc_fom, self.trial_vec, chunksize=self.opt.parallel_chunksize)

    def calc_trial_fom_parallel_mpi(self):
        """ Function to calculate the fom in parallel using mpi
        """
        step_len=int(len(self.trial_vec)/size)
        remain=int(len(self.trial_vec)%size)
        left, right=0, 0

        if rank<=remain-1:
            left=rank*(step_len+1)
            right=(rank+1)*(step_len+1)-1
        elif rank>remain-1:
            left= remain*(step_len+1)+(rank-remain)*step_len
            right= remain*(step_len+1)+(rank-remain+1)*step_len-1
        fom_temp=[]

        for i in range(left, right+1):
            fom_temp.append(parallel_calc_fom(self.trial_vec[i]))

        self.trial_fom=fom_temp

    # noinspection PyArgumentList
    def calc_error_bar(self, index):
        '''
        Calculates the errorbar for one parameter number index. 
        returns a float tuple with the error bars. fom_level is the 
        level which is the upperboundary of the fom is allowed for the
        calculated error.
        '''
        fom_level=self.opt.errorbar_level
        if self.setup_ok:  # and len(self.par_evals) != 0:
            par_values=self.par_evals[:, index]
            values_under_level=compress(self.fom_evals[:]<fom_level*self.best_fom, par_values)
            error_bar_low=values_under_level.min()-self.best_vec[index]
            error_bar_high=values_under_level.max()-self.best_vec[index]
            return error_bar_low, error_bar_high
        else:
            raise ErrorBarError()

    def init_new_generation(self, gen):
        ''' Function that is called every time a new generation starts'''
        pass

    def standard_init_new_generation(self, gen):
        ''' Function that is called every time a new generation starts'''
        pass

    def standard_update_pop(self, index):
        '''
        Function to update population vector index. calcs the figure of merit
        and compares it to the current population vector and also checks
        if it is better than the current best.
        '''
        # fom = self.calc_fom(self.trial_vec[index])
        fom=self.trial_fom[index]
        if fom<self.fom_vec[index]:
            self.pop_vec[index]=self.trial_vec[index].copy()
            self.fom_vec[index]=fom
            if fom<self.best_fom:
                self.new_best=True
                self.best_vec=self.trial_vec[index].copy()
                self.best_fom=fom

    # noinspection PyArgumentList
    def simplex_old_init_new_generation(self, gen):
        '''It will run the simplex method every simplex_interval
             generation with a fracitonal step given by simple_step 
             on the best indivual as well a random fraction of simplex_n individuals.
        '''
        iprint('Inits new generation')
        if gen%self.simplex_interval==0:
            spread=array(self.trial_vec).max(0)-array(self.trial_vec).min(0)
            simp=Simplex(self.calc_fom, self.best_vec, spread*self.simplex_step)
            iprint('Starting simplex run for best vec')
            new_vec, err, _iter=simp.minimize(epsilon=self.best_fom/self.simplex_rel_epsilon,
                                              maxiters=self.simplex_max_iter)
            iprint('FOM improvement: ', self.best_fom-err)

            if self.opt.use_boundaries:
                # Check so that the parameters lie inside the bounds
                ok=bitwise_and(self.par_max>new_vec, self.par_min<new_vec)
                # If not inside make a random re-initialization of that parameter
                new_vec=where(ok, new_vec, random.rand(self.n_dim)* \
                              (self.par_max-self.par_min)+self.par_min)

            new_fom=self.calc_fom(new_vec)
            if new_fom<self.best_fom:
                self.best_fom=new_fom
                self.best_vec=new_vec
                self.pop_vec[0]=new_vec
                self.fom_vec[0]=self.best_fom
                self.new_best=True

            # Apply the simplex to a simplex_n members (0-1)
            for index1 in random_mod.sample(range(len(self.pop_vec)),
                                            int(len(self.pop_vec)*self.simplex_n)):
                iprint('Starting simplex run for member: ', index1)
                mem=self.pop_vec[index1]
                mem_fom=self.fom_vec[index1]
                simp=Simplex(self.calc_fom, mem, spread*self.simplex_step)
                new_vec, err, _iter=simp.minimize(epsilon=self.best_fom/self.simplex_rel_epsilon,
                                                  maxiters=self.simplex_max_iter)
                if self.opt.use_boundaries:
                    # Check so that the parameters lie inside the bounds
                    ok=bitwise_and(self.par_max>new_vec, self.par_min<new_vec)
                    # If not inside make a random re-initialization of that parameter
                    new_vec=where(ok, new_vec, random.rand(self.n_dim)* \
                                  (self.par_max-self.par_min)+self.par_min)

                new_fom=self.calc_fom(new_vec)
                if new_fom<mem_fom:
                    self.pop_vec[index1]=new_vec
                    self.fom_vec[index1]=new_fom
                    if new_fom<self.best_fom:
                        self.best_fom=new_fom
                        self.best_vec=new_vec
                        self.new_best=True

    def simplex_init_new_generation(self, gen:int):
        '''
        It will run the simplex method every simplex_interval
        generation with a fracitonal step given by simple_step
        on the simplex_n*n_pop best individuals.
        '''
        iprint('Inits new generation')
        if gen%self.simplex_interval==0:
            # noinspection PyArgumentList
            spread=array(self.trial_vec).max(axis=0)-array(self.trial_vec).min(axis=0)

            idxs=argsort(self.fom_vec)
            n_ind=int(self.n_pop*self.simplex_n)
            if n_ind==0:
                n_ind=1
            # Apply the simplex to a simplex_n members (0-1)
            for index1 in idxs[:n_ind]:
                self.text_output('Starting simplex run for member: %d'%index1)
                mem=self.pop_vec[index1].copy()
                mem_fom=self.fom_vec[index1]
                simp=Simplex(self.calc_fom, mem, spread*self.simplex_step)
                new_vec, err, _iter=simp.minimize(epsilon=self.best_fom/self.simplex_rel_epsilon,
                                                  maxiters=self.simplex_max_iter)
                if self.opt.use_boundaries:
                    # Check so that the parameters lie inside the bounds
                    ok=bitwise_and(self.par_max>new_vec, self.par_min<new_vec)
                    # If not inside make a random re-initialization of that parameter
                    new_vec=where(ok, new_vec, random.rand(self.n_dim)* \
                                  (self.par_max-self.par_min)+self.par_min)

                new_fom=self.calc_fom(new_vec)
                if new_fom<mem_fom:
                    self.pop_vec[index1]=new_vec.copy()
                    self.fom_vec[index1]=new_fom
                    if new_fom<self.best_fom:
                        self.best_fom=new_fom
                        self.best_vec=new_vec.copy()
                        self.new_best=True

    def simplex_best_1_bin(self, index):
        return self.best_1_bin(index)

    def jade_update_pop(self, index):
        '''
        A modified update pop to handle the JADE variation of Differential evolution
        '''
        fom=self.trial_fom[index]
        if fom<self.fom_vec[index]:
            self.pop_vec[index]=self.trial_vec[index].copy()
            self.fom_vec[index]=fom
            self.updated_kr.append(self.kr_vec[index])
            self.updated_km.append(self.km_vec[index])
            if fom<self.best_fom:
                self.new_best=True
                self.best_vec=self.trial_vec[index].copy()
                self.best_fom=fom

    def jade_init_new_generation(self, gen):
        '''
        A modified generation update for jade
        '''
        if gen>1:
            updated_kms=array(self.updated_km)
            updated_krs=array(self.updated_kr)
            if len(updated_kms)!=0:
                self.opt.km=(1.0-self.c)*self.opt.km+self.c*sum(updated_kms**2)/sum(updated_kms)
                self.opt.kr=(1.0-self.c)*self.opt.kr+self.c*mean(updated_krs)
        self.km_vec=abs(self.opt.km+random.standard_cauchy(self.n_pop)*0.1)
        self.kr_vec=self.opt.kr+random.normal(size=self.n_pop)*0.1
        iprint('km: ', self.opt.km, ', kr: ', self.opt.kr)
        self.km_vec=where(self.km_vec>0, self.km_vec, 0)
        self.km_vec=where(self.km_vec<1, self.km_vec, 1)
        self.kr_vec=where(self.kr_vec>0, self.kr_vec, 0)
        self.kr_vec=where(self.kr_vec<1, self.kr_vec, 1)

        self.updated_kr=[]
        self.updated_km=[]

    def jade_best(self, index):
        vec=self.pop_vec[index]
        # Create mutation vector
        # Select two random vectors for the mutation
        index1=int(random.rand(1)*self.n_pop)
        index2=int(random.rand(1)*len(self.par_evals))
        # Make sure it is not the same vector 
        # while index2 == index1:
        #    index2 = int(random.rand(1)*self.n_pop)

        # Calculate the mutation vector according to the best/1 scheme
        mut_vec=vec+self.km_vec[index]*(self.best_vec-vec)+self.km_vec[index]*(
                self.pop_vec[index1]-self.par_evals[index2])

        # Binomial test to determine which parameters to change
        # given by the recombination constant kr
        recombine=random.rand(self.n_dim)<self.kr_vec[index]
        # Make sure at least one parameter is changed
        recombine[int(random.rand(1)*self.n_dim)]=1
        # Make the recombination
        trial=where(recombine, mut_vec, vec)

        # Implementation of constrained optimization
        if self.opt.use_boundaries:
            # Check so that the parameters lie inside the bounds
            ok=bitwise_and(self.par_max>trial, self.par_min<trial)
            # If not inside make a random re-initialization of that parameter
            trial=where(ok, trial, random.rand(self.n_dim)* \
                        (self.par_max-self.par_min)+self.par_min)
        self.trial_vec[index]=trial
        # return trial

    def best_1_bin(self, index):
        '''
        The default create_trial function for this class. 
        uses the best1bin method to create a new vector from the population.
        '''
        vec=self.pop_vec[index]
        # Create mutation vector
        # Select two random vectors for the mutation
        index1=int(random.rand(1)*self.n_pop)
        index2=int(random.rand(1)*self.n_pop)
        # Make sure it is not the same vector 
        while index2==index1:
            index2=int(random.rand(1)*self.n_pop)

        # Calculate the mutation vector according to the best/1 scheme
        mut_vec=self.best_vec+self.opt.km*(
                self.pop_vec[index1]-self.pop_vec[index2])

        # Binomial test to determine which parameters to change
        # given by the recombination constant kr
        recombine=random.rand(self.n_dim)<self.opt.kr
        # Make sure at least one parameter is changed
        recombine[int(random.rand(1)*self.n_dim)]=1
        # Make the recombination
        trial=where(recombine, mut_vec, vec)

        # Implementation of constrained optimization
        if self.opt.use_boundaries:
            # Check so that the parameters lie inside the bounds
            ok=bitwise_and(self.par_max>trial, self.par_min<trial)
            # If not inside make a random re-initialization of that parameter
            trial=where(ok, trial, random.rand(self.n_dim)* \
                        (self.par_max-self.par_min)+self.par_min)

        self.trial_vec[index]=trial
        # return trial

    def best_either_or(self, index):
        '''
        The either/or scheme for creating a trial. Using the best vector
        as base vector.
        '''
        vec=self.pop_vec[index]
        # Create mutation vector
        # Select two random vectors for the mutation
        index1=int(random.rand(1)*self.n_pop)
        index2=int(random.rand(1)*self.n_pop)
        # Make sure it is not the same vector 
        while index2==index1:
            index2=int(random.rand(1)*self.n_pop)

        if random.rand(1)<self.pf:
            # Calculate the mutation vector according to the best/1 scheme
            trial=self.best_vec+self.opt.km*(
                    self.pop_vec[index1]-self.pop_vec[index2])
        else:
            # Trying something else out more like normal recombination
            trial=vec+self.opt.kr*(
                    self.pop_vec[index1]+self.pop_vec[index2]-2*vec)

        # Implementation of constrained optimization
        if self.opt.use_boundaries:
            # Check so that the parameters lie inside the bounds
            ok=bitwise_and(self.par_max>trial, self.par_min<trial)
            # If not inside make a random re-initialization of that parameter
            trial=where(ok, trial, random.rand(self.n_dim)* \
                        (self.par_max-self.par_min)+self.par_min)
        self.trial_vec[index]=trial
        # return trial

    def rand_1_bin(self, index):
        '''
        The default create_trial function for this class. 
        uses the best1bin method to create a new vector from the population.
        '''
        vec=self.pop_vec[index]
        # Create mutation vector
        # Select three random vectors for the mutation
        index1=int(random.rand(1)*self.n_pop)
        index2=int(random.rand(1)*self.n_pop)
        # Make sure it is not the same vector 
        while index2==index1:
            index2=int(random.rand(1)*self.n_pop)
        index3=int(random.rand(1)*self.n_pop)
        while index3==index1 or index3==index2:
            index3=int(random.rand(1)*self.n_pop)

        # Calculate the mutation vector according to the rand/1 scheme
        mut_vec=self.pop_vec[index3]+self.opt.km*(
                self.pop_vec[index1]-self.pop_vec[index2])

        # Binomial test to determine which parameters to change
        # given by the recombination constant kr
        recombine=random.rand(self.n_dim)<self.opt.kr
        # Make sure at least one parameter is changed
        recombine[int(random.rand(1)*self.n_dim)]=1
        # Make the recombination
        trial=where(recombine, mut_vec, vec)

        # Implementation of constrained optimization
        if self.opt.use_boundaries:
            # Check so that the parameters lie inside the bounds
            ok=bitwise_and(self.par_max>trial, self.par_min<trial)
            # If not inside make a random re-initialization of that parameter
            trial=where(ok, trial, random.rand(self.n_dim)* \
                        (self.par_max-self.par_min)+self.par_min)
        self.trial_vec[index]=trial
        # return trial

    def rand_either_or(self, index):
        '''
        random base vector either/or trial scheme
        '''
        # Create mutation vector
        # Select two random vectors for the mutation
        index1=int(random.rand(1)*self.n_pop)
        index2=int(random.rand(1)*self.n_pop)
        # Make sure it is not the same vector 
        while index2==index1:
            index2=int(random.rand(1)*self.n_pop)
        index0=int(random.rand(1)*self.n_pop)
        while index0==index1 or index0==index2:
            index0=int(random.rand(1)*self.n_pop)

        if random.rand(1)<self.pf:
            # Calculate the mutation vector according to the best/1 scheme
            trial=self.pop_vec[index0]+self.opt.km*(
                    self.pop_vec[index1]-self.pop_vec[index2])
        else:
            # Calculate a continuous recombination
            # Trying something else out more like normal recombination
            trial=self.pop_vec[index0]+self.opt.kr*(
                    self.pop_vec[index1]+self.pop_vec[index2]-2*self.pop_vec[index0])

        # Implementation of constrained optimization
        if self.opt.use_boundaries:
            # Check so that the parameters lie inside the bounds
            ok=bitwise_and(self.par_max>trial, self.par_min<trial)
            # If not inside make a random re-initialization of that parameter
            trial=where(ok, trial, random.rand(self.n_dim)* \
                        (self.par_max-self.par_min)+self.par_min)
        self.trial_vec[index]=trial
        # return trial

    # Different function for accessing and setting parameters that
    # the user should have control over.
    def plot_output(self):
        data=SolverUpdateInfo(
            fom_value=self.model.fom,
            fom_name=self.model.fom_func.__name__,
            fom_log=self.get_fom_log(),
            new_best=self.new_best,
            data=self.model.data
            )
        self._callbacks.plot_output(data)

    def text_output(self, text: str):
        self._callbacks.text_output(text)

    def parameter_output(self):
        param_info=SolverParameterInfo(values=self.best_vec.copy(),
                                       new_best=self.new_best,
                                       population=[vec.copy() for vec in self.pop_vec],
                                       max_val=self.par_max.copy(),
                                       min_val=self.par_min.copy(),
                                       fitting=True)
        self._callbacks.parameter_output(param_info)

    def autosave(self):
        self._callbacks.autosave()

    def fitting_ended(self):
        result = self.get_result_info()
        self._callbacks.fitting_ended(result)

    def get_result_info(self):
        result = SolverResultInfo(
            start_guess=self.start_guess.copy(),
            error_message=self.error,
            values=self.best_vec.copy(),
            new_best=self.new_best,
            population=[vec.copy() for vec in self.pop_vec],
            max_val=self.par_max.copy(),
            min_val=self.par_min.copy(),
            fitting=True
            )
        return result

    def set_callbacks(self, callbacks: GenxOptimizerCallback):
        self._callbacks=callbacks

    def get_callbacks(self) -> 'GenxOptimizerCallback':
        return self._callbacks

    def get_model(self):
        '''
        Getter that returns the model in use in solver.
        '''
        return self.model

    def get_fom_log(self):
        '''
        Returns the fom as a fcn of iteration in an array. 
        Last element last fom value
        '''
        return array(self.fom_log)

    def get_create_trial(self, index=False):
        '''
        returns the current create trial function name if index is False as
        a string or as index in the mutation_schemes list.
        '''
        pos=self.mutation_schemes.index(self.create_trial)
        if index:
            # return the position
            return pos
        else:
            # return the name
            return self.mutation_schemes[pos].__name__

    def set_km(self, val):
        self.opt.km=val

    def set_kr(self, val):
        self.opt.kr=val

    def set_create_trial(self, val):
        '''
        Raises LookupError if the value val [string] does not correspond
        to a mutation scheme/trial function
        '''
        # Get the names of the available functions
        names=[f.__name__ for f in self.mutation_schemes]
        # Find the position of val

        pos=names.index(val)
        self.create_trial=self.mutation_schemes[pos]
        self.opt.create_trial=val
        if val=='jade_best':
            self.update_pop=self.jade_update_pop
            self.init_new_generation=self.jade_init_new_generation
        elif val=='simplex_best_1_bin':
            self.init_new_generation=self.simplex_init_new_generation
            self.update_pop=self.standard_update_pop
        else:
            self.init_new_generation=self.standard_init_new_generation
            self.update_pop=self.standard_update_pop

    def set_pop_mult(self, val):
        self.opt.pop_mult=val

    def set_pop_size(self, val):
        self.opt.pop_size=int(val)

    def set_max_generations(self, val):
        self.opt.max_generations=int(val)

    def set_max_generation_mult(self, val):
        self.opt.max_generation_mult=val

    def set_sleep_time(self, val):
        self.opt.sleep_time=val

    def set_max_log(self, val):
        self.opt.max_log_elements=val

    def set_use_pop_mult(self, val):
        self.opt.use_pop_mult=val

    def set_use_max_generations(self, val):
        self.opt.use_max_generations=val

    def set_use_start_guess(self, val):
        self.opt.use_start_guess=val

    def set_use_boundaries(self, val):
        self.opt.use_boundaries=val

    def set_use_autosave(self, val):
        self.opt.use_autosave=val

    def set_autosave_interval(self, val):
        self.opt.autosave_interval=int(val)

    def set_use_parallel_processing(self, val):
        if __parallel_loaded__:
            self.opt.use_parallel_processing=val
            self.opt.use_mpi=False if val else self.opt.use_mpi
        else:
            self.opt.use_parallel_processing=False

    def set_use_mpi(self, val):
        """Sets if mpi should use for parallel optimization"""
        if __mpi_loaded__:
            self.opt.use_mpi=val
            self.opt.use_parallel_processing=False if val else self.opt.use_parallel_processing
        else:
            self.opt.use_mpi=False

    def set_processes(self, val):
        self.opt.parallel_processes=int(val)

    def set_chunksize(self, val):
        self.opt.parallel_chunksize=int(val)

    def set_fom_allowed_dis(self, val):
        self.opt.allowed_fom_discrepancy=float(val)

# ==============================================================================
# Functions that is needed for parallel processing!
model=Model(); par_funcs=() # global variables set in functions below

def parallel_init(model_copy: Model, numba_procs=None, use_mpi=False):
    '''
    parallel initialization of a pool of processes. The function takes a
    pickle safe copy of the model and resets the script module and the compiles
    the script and creates function to set the variables.
    '''
    if not use_mpi:
        # ignore KeyboardInterrupt so that master process can handle it
        import signal
        signal.signal(signal.SIGINT, signal.SIG_IGN)

    if numba_procs is not None:
        import numba
        if hasattr(numba, 'set_num_threads'):
            iprint(f"Setting numba threads to {numba_procs}")
            numba.set_num_threads(numba_procs)
    global model, par_funcs
    model=model_copy
    model.reset()
    model.simulate()
    (par_funcs, start_guess, par_min, par_max)=model.get_fit_pars()

def init_cuda():
    iprint("Init CUDA in one worker")
    # activate cuda in subprocesses
    from .models.lib import paratt_cuda
    from .models.lib import neutron_cuda
    from .models.lib import paratt, neutron_refl
    paratt.Refl=paratt_cuda.Refl
    paratt.ReflQ=paratt_cuda.ReflQ
    paratt.Refl_nvary2=paratt_cuda.Refl_nvary2
    neutron_refl.Refl=neutron_cuda.Refl
    from .models.lib import paratt, neutron_refl
    paratt.Refl=paratt_cuda.Refl
    paratt.ReflQ=paratt_cuda.ReflQ
    paratt.Refl_nvary2=paratt_cuda.Refl_nvary2
    neutron_refl.Refl=neutron_cuda.Refl
    iprint("CUDA init done, go to work")


def parallel_calc_fom(vec):
    '''
    function that is used to calculate the fom in a parallel process.
    It is a copy of calc_fom in the DiffEv class
    '''
    global model, par_funcs
    # set the parameter values in the model
    list(map(lambda func, value: func(value), par_funcs, vec))
    # evaluate the model and calculate the fom
    fom=model.evaluate_fit_func()

    return fom

def _calc_fom(model_obj: Model, vec, param_funcs):
    '''
    Function to calcuate the figure of merit for parameter vector
    vec.
    '''
    # Set the parameter values
    list(map(lambda func, value: func(value), param_funcs, vec))

    return model_obj.evaluate_fit_func()

class CircBuffer:
    '''A buffer with a fixed length to store the logging data from the diffev 
    class. Initilized to a maximumlength after which it starts to overwrite
    the data again.
    '''

    def __init__(self, maxlen, buffer=None):
        '''Inits the class with a certain maximum length maxlen.
        '''
        self.maxlen=int(maxlen)
        self.pos=-1
        self.filled=False
        if buffer is None:
            self.buffer=zeros((self.maxlen,))
        else:
            if len(buffer)!=0:
                self.buffer=array(buffer).repeat(
                    ceil(self.maxlen/(len(buffer)*1.0)), 0)[:self.maxlen]
                self.pos=len(buffer)-1
            else:
                self.buffer=zeros((self.maxlen,)+buffer.shape[1:])

    def reset(self, buffer=None):
        '''Resets the buffer to the initial state
        '''
        self.pos=-1
        self.filled=False
        # self.buffer = buffer
        if buffer is None:
            self.buffer=zeros((self.maxlen,))
        else:
            if len(buffer)!=0:
                self.buffer=array(buffer).repeat(
                    ceil(self.maxlen/(len(buffer)*1.0)), 0)[:self.maxlen]
                self.pos=len(buffer)-1
            else:
                self.buffer=zeros((self.maxlen,)+buffer.shape[1:])

    def append(self, item, axis=None):
        '''Appends an element to the last position of the buffer
        '''
        new_pos=(self.pos+1)%self.maxlen
        if len(self.buffer)>=self.maxlen:
            if self.pos>=(self.maxlen-1):
                self.filled=True
            self.buffer[new_pos]=array(item).real
        else:
            self.buffer=append(self.buffer, item, axis=axis)
        self.pos=new_pos

    def array(self):
        '''returns an ordered array instead of the circular
        working version
        '''
        if self.filled:
            return r_[self.buffer[self.pos+1:], self.buffer[:self.pos+1]]
        else:
            return r_[self.buffer[:self.pos+1]]

    def copy_from(self, other):
        '''Add copy support
        '''
        if type(other)==type(array([])):
            self.buffer= other[-self.maxlen:]
            self.pos=len(self.buffer)-1
            self.filled=self.pos>=(self.maxlen-1)
        elif other.__class__==self.__class__:
            # Check if the buffer has been removed.
            if len(other.buffer)==0:
                self.__init__(other.maxlen, other.buffer)
            else:
                self.buffer=other.buffer.copy()
                self.maxlen=other.maxlen
                self.pos=other.pos
                try:
                    self.filled=other.filled
                except AttributeError:
                    self.filled=False
        else:
            raise TypeError('CircBuffer support only copying from CircBuffer'
                            ' and arrays.')

    def __len__(self):
        if self.filled:
            return len(self.buffer)
        else:
            return (self.pos>0)*self.pos

    def __getitem__(self, item):
        return self.array().__getitem__(item)



'''
Module for usefull decorators e.g. for logging function calls, input and output.
'''

import inspect
import logging
import threading
from time import time
from io import StringIO
from functools import update_wrapper

def getinfo(func):
    """
      Returns an info dictionary containing:
      - name (the name of the function : str)
      - argnames (the names of the arguments : list)
      - defaults (the values of the default arguments : tuple)
      - signature (the signature : str)
      - doc (the docstring : str)
      - module (the module name : str)
      - dict (the function __dict__ : str)
    """
    assert inspect.ismethod(func) or inspect.isfunction(func)
    signature=inspect.signature(func)
    output=dict(name=func.__name__, argnames=list(signature.parameters.keys()),
                signature=str(signature).lstrip('(').rstrip(')'),
                defaults=func.__defaults__, doc=func.__doc__,
                module=func.__module__, dict=func.__dict__,
                globals=func.__globals__, closure=func.__closure__)
    return output

# the real meat is here
def _decorator(caller, func):
    infodict=getinfo(func)
    argnames=infodict['argnames']
    assert not ('_call_' in argnames or '_func_' in argnames), \
        'You cannot use _call_ or _func_ as argument names!'
    src="lambda %(signature)s: _call_(_func_, %(signature)s)"%infodict
    dec_func=eval(src, dict(_func_=func, _call_=caller))
    return update_wrapper(dec_func, func)

def decorator(caller):
    """
    General purpose decorator factory: takes a caller function as
    input and returns a decorator with the same attributes.
    """
    return update_wrapper(lambda f: _decorator(caller, f), caller)

########################### Decorators for logging ##################################

def _nicetype(item):
    if 'array' in type(item).__name__:
        return type(item).__name__+("[%s]"%str(item.shape))
    elif type(item) in [list, tuple]:
        return type(item).__name__+('[%i]'%len(item))
    else:
        return type(item).__name__

class DecoratorLogger(logging.getLoggerClass()):
    '''
      A logger that makes sure the actual function definition filename, lineno and function name
      is used for logging.
    '''

    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
        if extra is None:
            return logging.getLoggerClass().makeRecord(self, name, level, fn, lno,
                                                       msg, args, exc_info, func=func, extra=None)
        else:
            return logging.getLoggerClass().makeRecord(self, name, level, extra['name'], extra['lno'],
                                                       msg, args, exc_info, func=extra['func'], extra=None)

# only use the debug decorators in debugging runs.
old_class=logging.getLoggerClass()
logging.setLoggerClass(DecoratorLogger)
logger=logging.getLogger('deco')
logger.setLevel(logging.DEBUG)
null_handler=logging.StreamHandler(StringIO())
null_handler.setLevel(logging.CRITICAL)
logger.addHandler(null_handler)
logging.setLoggerClass(old_class)

def _logformat(msg, decname, func):
    if hasattr(func, 'im_func'):
        co=func.__func__.__code__
    else:
        co=func.__code__
    fname=co.co_filename
    lno=co.co_firstlineno
    logger.debug(msg, extra={'func': '@'+decname, 'name': fname, 'lno': lno})

@decorator
def log_call(func, *args, **kw):
    '''
      Decorator to log just the method call.
    '''
    if logging.root.getEffectiveLevel()>logging.DEBUG: return func(*args, **kw)
    infodict=getinfo(func)
    if len(infodict['argnames'])>0 and infodict['argnames'][0]=='self':
        _logformat('%s.%s.%s <%s>'%(infodict['module'],
                                    args[0].__class__.__name__, infodict['name'], repr(threading.current_thread())),
                   'log_call', func)
    else:
        _logformat('%s.%s <%s>'%(infodict['module'], infodict['name'], repr(threading.current_thread())), 'log_call',
                   func)
    return func(*args, **kw)

@decorator
def log_input(func, *args, **kw):
    '''
      Decorator to log a method call with input.
    '''
    if logging.root.getEffectiveLevel()>logging.DEBUG: return func(*args, **kw)
    infodict=getinfo(func)
    if hasattr(func, 'im_func'):
        logstr=' call %s.%s('%(args[0].__class__.__name__, infodict['name'])
        method=True
        info_len=len('%s.%s('%(args[0].__class__.__name__, infodict['name']))
    else:
        logstr='call %s('%(infodict['name'])
        method=False
        info_len=len('%s('%(infodict['name']))

    namelen=max([len(name)+3 for name in infodict['argnames']]+[len(name) for name in list(kw.keys())])
    for i, arg in enumerate(args):
        if i==0 and (method or infodict['name']=='__init__'):
            continue
        value=repr(arg)
        value_split=value.splitlines()
        if len(value_split)>5:
            value="\n".join(value_split[:2]+[' '*int(len(value_split[1])/2)+'...']+value_split[-2:])
        value.replace('\n', '\n'+' '*int(namelen+7))
        logstr+='\n'+('%% %is= %%s (%%s)'%(namelen+3))%(infodict['argnames'][i], value, _nicetype(arg))

    for key, arg in sorted(kw.items()):
        value=repr(arg)
        value_split=value.splitlines()
        if len(value_split)>5:
            value="\n".join(value_split[:2]+[' '*int(len(value_split[1])/2)+'...']+value_split[-2:])
        value.replace('\n', '\n'+' '*int(namelen+4))
        logstr+='\n'+('KW %% %ss= %%s (%%s)'%namelen)%(key, repr(value), _nicetype(arg))
    logstr+='\n)'
    logstr=logstr.replace('\n', '\n'+' '*int(info_len+10))
    logstr=logstr.replace('call ', 'call \n'+' '*10)

    _logformat(logstr, 'log_input', func)
    return func(*args, **kw)

@decorator
def log_output(func, *args, **kw):
    '''
      Decorator to log a method call with output. If combined with log_input
      the input is logged at the time before the call and the output after.
    '''
    if logging.root.getEffectiveLevel()>logging.DEBUG: return func(*args, **kw)
    output=func(*args, **kw)
    infodict=getinfo(func)
    if len(infodict['argnames'])>0 and infodict['argnames'][0]=='self':
        logstr='return from %s.%s'%(args[0].__class__.__name__, infodict['name'])
    else:
        logstr='return from %s'%(infodict['name'])
    value=repr(output)
    value_split=value.splitlines()
    if len(value_split)>5:
        value="\n".join(value_split[:2]+[' '*int(len(value_split[1])/2)+'...']+value_split[-2:])
    logstr+='\n-> %15s (%s)'%(value, _nicetype(output))
    logstr=logstr.replace('\n', '\n'+' '*44)
    _logformat(logstr, 'log_output', func)
    return output

@decorator
def log_both(func, *args, **kw):
    '''
      Decoratore to log a method call with input and output.
    '''
    if logging.root.getEffectiveLevel()>logging.DEBUG: return func(*args, **kw)
    infodict=getinfo(func)
    if hasattr(func, 'im_func'):
        logstr=' call %s.%s('%(args[0].__class__.__name__, infodict['name'])
        method=True
        info_len=len('%s.%s('%(args[0].__class__.__name__, infodict['name']))
    else:
        logstr='call %s('%(infodict['name'])
        method=False
        info_len=len('%s('%(infodict['name']))

    namelen=max([len(name)+3 for name in infodict['argnames']]+[len(name) for name in list(kw.keys())])
    for i, arg in enumerate(args):
        if i==0 and (method or infodict['name']=='__init__'):
            continue
        value=repr(arg)
        value_split=value.splitlines()
        if len(value_split)>5:
            value="\n".join(value_split[:2]+[' '*int(len(value_split[1])/2)+'...']+value_split[-2:])
        value.replace('\n', '\n'+' '*int(namelen+7))
        logstr+='\n'+('%% %is= %%s (%%s)'%(namelen+3))%(infodict['argnames'][i], value, _nicetype(arg))

    for key, arg in sorted(kw.items()):
        value=repr(arg)
        value_split=value.splitlines()
        if len(value_split)>5:
            value="\n".join(value_split[:2]+[' '*int(len(value_split[1])/2)+'...']+value_split[-2:])
        value.replace('\n', '\n'+' '*int(namelen+4))
        logstr+='\n'+('KW %% %ss= %%s (%%s)'%namelen)%(key, repr(value), _nicetype(arg))
    logstr+='\n)'
    logstr=logstr.replace('\n', '\n'+' '*int(info_len+10))
    logstr=logstr.replace('call ', 'call \n'+' '*10)

    _logformat(logstr, 'log_input', func)
    # call the function
    output=func(*args, **kw)
    if len(infodict['argnames'])>0 and infodict['argnames'][0]=='self':
        logstr='return from %s.%s'%(args[0].__class__.__name__, infodict['name'])
    else:
        logstr='return from %s'%(infodict['name'])
    value=repr(output)
    value_split=value.splitlines()
    if len(value_split)>5:
        value="\n".join(value_split[:2]+[' '*int(len(value_split[1])/2)+'...']+value_split[-2:])
    logstr+='\n-> %s (%s)'%(value, _nicetype(output))
    logstr=logstr.replace('\n', '\n'+' '*int(info_len+10))
    _logformat(logstr, 'log_output', func)
    return output

timings={}

@decorator
def time_call(func, *args, **kw):
    '''
      Decorator to log just the method call.
    '''
    if logging.root.getEffectiveLevel()>logging.DEBUG: return func(*args, **kw)
    name=func.__name__
    start=time()
    output=func(*args, **kw)
    runtime=time()-start
    if not name in timings:
        timings[name]=(0., 0.)
    avg, calls=timings[name]
    timings[name]=((avg*calls+runtime)/(calls+1), calls+1)
    return output

########################## General decorators ###############################

class check_input(object):
    '''
      Decorator checking the input to a function.

      Usage:
      @check_input([type1, type2, ...], try_convert=True)
      def your_function(var1, var2, ...):
          ... function text ...
    '''
    try_convert=True
    types=[]

    def __init__(self, types, try_convert=True):
        self.types=types
        self.try_convert=try_convert

    def __call__(self, func):
        infodict=getinfo(func)
        argnames=infodict['argnames']
        assert not ('_call_' in argnames or '_func_' in argnames), \
            'You cannot use _call_ or _func_ as argument names!'
        src="def dec_func(%(signature)s):"%infodict
        for i, typei in enumerate(self.types):
            src+='\n  if type(%s).__name__!="%s":'%(argnames[i], typei.__name__)
            if self.try_convert:
                src+='\n    try:'
                src+='\n      %s=%s(%s)'%(argnames[i], typei.__name__, argnames[i])
                src+='\n    except:'
                src+='\n      raise ValueError("type of %s is not %s")'%(argnames[i], typei.__name__)
            else:
                src+='\n    raise ValueError("type of %s is not %s")'%(argnames[i], typei.__name__)
        src+='\n  return _func_(%(signature)s)'%infodict
        exec_dict=dict(_func_=func, _call_=self.__call__)
        exec(src, exec_dict)
        return update_wrapper(exec_dict['dec_func'], func)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
.. module:: grating
   :synopsis: Models to calcualte the reflectivity from a grating

.. moduleauthor:: Matts Björck <matts.bjorck@gmail.com>

This module provide functions to calculate the reflectivity from gratings using the distorted wave Born approximation
(DWBA) and the kinematical theory.
"""

import numpy as np
from scipy.special import j1
from functools import reduce
from genx.core.custom_logging import iprint

_ctype=np.complex128

def dot2(A, B):
    D=np.zeros(A.shape, dtype=_ctype)
    D[0, 0]=A[0, 0]*B[0, 0]+A[0, 1]*B[1, 0]
    D[0, 1]=A[0, 0]*B[0, 1]+A[0, 1]*B[1, 1]
    D[1, 0]=A[1, 0]*B[0, 0]+A[1, 1]*B[1, 0]
    D[1, 1]=A[1, 0]*B[0, 1]+A[1, 1]*B[1, 1]
    return D

def ass_P(k, d, ctype=_ctype):
    """ Assemble the propagation matrix for the layers"""
    P=np.zeros((2, 2)+k.shape, dtype=ctype)
    P[0, 0]=np.exp(-1.0J*k*d)
    P[1, 1]=1.0/P[0, 0]
    return P

def ass_X(k, ctype=_ctype):
    """Assemble the interface matrix for all interfaces"""
    # k_j = k[..., 1:]
    # k_jm1 = k[..., :-1]
    k_j=k[..., :-1]
    k_jm1=k[..., 1:]

    X=np.empty((2, 2)+k_j.shape, dtype=ctype)

    X[0, 0]=(k_j+k_jm1)/k_jm1/2
    X[0, 1]=(k_jm1-k_j)/k_jm1/2
    X[1, 0]=X[0, 1]
    X[1, 1]=X[0, 0]
    return X

def calc_fields(kz, lamda, v, d, return_sub=False):
    """ Calculate the fields inside a multilayer"""
    k_vac=2*np.pi/lamda
    kz=kz[:, np.newaxis]
    v_amb=v[-1]
    k_j=np.sqrt(kz*kz-v+v_amb*(1-(kz/k_vac)**2))

    X=ass_X(k_j)
    P=ass_P(k_j, d)

    M=np.ones_like(X)
    # Iterate over all layers - starting at the bottom (index 0)
    M[..., 0]=X[..., 0]
    for i in range(1, X.shape[-1]):
        M[..., i]=dot2(dot2(X[..., i], P[..., i]), M[..., i-1])

    # Calculations of the wavefields above each interface
    E_T_sub=1./M[0, 0, :, -1][:, np.newaxis]
    E_T=M[0, 0]*E_T_sub
    E_R=M[1, 0]*E_T_sub

    # r = M[1, 0, :, -1] / M[0, 0, :, -1]
    # r = E_R[:, -1]/E_T[:, -1]
    if return_sub:
        return E_T, E_R, k_j, E_T_sub
    else:
        return E_T, E_R, k_j

def coherent_nonspec_refl(k_in, k_out, lamda, vf_layers, v_mean, d, kin=False, spec_refl=False):
    """Calculates the coherent matrix element T_12 for non-specular rods using DWBA or kinematical theory.

    Parameters:
        k_in (array): The wave vector of the incoming beam (k_x, k_y, k_z), size 3xM.
        k_out (array): The wave vector of the reflected beam (k_x, k_y, k_z), size 3xM.
        lamda (array): The wavelength of the radiation used, size M.
        vf_layers (array): The in-plane fourier transform function of scattering
                                potential for the layers, size MxN.
        v_mean (array): The mean scattering potential for the layers, size N.
        d (array): The thickness of layers. First layer is the substrate, size N.
        kin (bool): Flag for the use of kinematical theory (defualt: False)
        spec_refl (bool): Flag for the return of the specular reflected amplitude.

    Returns:
        T_12 (array): The scattering matrix element
    """
    # vf_func = assemble_vf_func(vf)
    # v_mean = vf_func(np.zeros(2))/area_uc
    T_1, R_1, kz_1=calc_fields(k_in[2], lamda, v_mean, d)
    T_2, R_2, kz_2=calc_fields(-k_out[2], lamda, v_mean, d)

    # Note that the ambient layer, index -1, and the substrate, index 0, is not used and must be removed.
    q0=-(kz_1+kz_2)[:, 1:-1]
    q1=-(kz_1-kz_2)[:, 1:-1]
    q2=-(-kz_1+kz_2)[:, 1:-1]
    q3=-(-kz_2-kz_1)[:, 1:-1]

    # The in-plane scattering vector - check that this is correct!!!!
    # q_ip = k_out[:2] - k_in[:2]
    d=d[np.newaxis, 1:-1]*np.ones_like(R_2)[:, :-1]
    # vf = vf_func(q_ip).T[:, 1:-1]
    vf_layers=vf_layers[:, 1:-1]
    # The wave amplitudes T_x, R_x does not contain the amplitudes in the substrate, therefore :-1
    w=vf_layers*(T_2[:, :-1]*T_1[:, :-1]*ftz_layer(q0, d)+
                 R_2[:, :-1]*T_1[:, :-1]*ftz_layer(q1, d)+
                 T_2[:, :-1]*R_1[:, :-1]*ftz_layer(q2, d)+
                 R_2[:, :-1]*R_1[:, :-1]*ftz_layer(q3, d)
                 )
    T_12=w.sum(axis=-1)

    if kin:
        q=(k_out[2]-k_in[2])
        z=d.cumsum(axis=1)
        z=z[:, -1][:, np.newaxis]-z+d
        w=vf_layers*ftz_layer(-q[:, np.newaxis], d)*np.exp(1.0J*q[:, np.newaxis]*z)
        T_12=w.sum(axis=-1)
        # Adding the reflectivity from the substrate if at the specular rod
        T_12+=np.where(is_spec(k_in, k_out), vf[:, 0]*1.0J/q*np.exp(1.0J*q*z[:, 0]), 0.0)

    if spec_refl:
        return T_12, R_1[:, -1]  # /T_1[:, -1]
    else:
        return T_12

def coherent_refl(k_in, k_out, lamda, vf, v_mean, d, area_uc, kin=False):
    """Calculates the coherent reflectivity using DWBA.

        Parameters:
            k_in (array): The wave vector of the incoming beam (k_x, k_y, k_z), size 3xM.
            k_out (array): The wave vector of the reflected beam (k_x, k_y, k_z), size 3xM.
            lamda (array): The wavelength of the radiation used, size M.
            vf (array): The in-plane fourier transform function of scattering
                                    potential for the layers, size MxN.
            v_mean(array): The mean scattering potential for the layers, size N
            d (array): The thickness of layers. First layer is the substrate, size N.
            area_uc (float): The area of the unit cell in  AA^2.

        Returns:
            T_12 (array): The scattering matrix element
    """
    T_12, r=coherent_nonspec_refl(k_in, k_out, lamda, vf, v_mean, d, kin=kin, spec_refl=True)
    R=np.where(np.bitwise_and(is_spec(k_in, k_out), not kin),
               np.abs(r)**2,
               1./np.abs(k_in[2])/np.abs(k_out[2])/4/area_uc**2*np.abs(T_12)**2
               )
    return R

def is_spec(k_in, k_out):
    return np.bitwise_and(np.isclose(k_in[0], k_out[0]),
                          np.bitwise_and(np.isclose(k_in[1], 0), np.isclose(k_out[1], 0)))

def ftz_layer(qz, d):
    """The one-diminesional fourier transform for a homogenous layer of thickness d"""
    return (1.0-np.exp(1.0J*qz*d))/-1.0J/qz

def assemble_vf_func(vf_list):
    """Creates a function to evaluate vf

    Parameters:
        vf_list (list): A list of vf functions, one for each layer

    Returns:
        vf_func (function): A function that evaluates the vf functions given q-coords
    """

    def vf_func(q):
        return np.array([vf(q) for vf in vf_list])

    return vf_func

# from refl import ReflFunction as ArithFunc
# This is a GenX dependency which is not strictly needed but makes the life easier
# def ft_math(func):
#    """Decorator that enables a fourier transform function to used in arithmetic expressions"""
#    return ArithFunc(func, np.zeros(2), {}, id='ft(q)')


def mult(factor, func):
    """Multilies factor with func"""

    def mult_func(*args, **kwargs):
        return factor*func(*args, **kwargs)

    return mult_func

def ft_circle(radius=1.):
    """Creates a function that evaluates the fourier transform of a circle with radius"""

    # @ft_math
    def circle(q):
        q_r=np.sqrt(q[0]**2+q[1]**2)
        return 2*np.pi*radius**2*jinc(q_r*radius)

    return circle

def ft_layer(area_uc):
    """Creates a function that returns the in-plane fourier transform of an in-plane infinite layer"""

    # @ft_math
    def layer(q):
        return np.where(np.bitwise_and(q[0]==0, q[1]==0), 1.0, 0.0)*area_uc

    return layer

def jinc(x):
    """The jinc function J_1(x)/x, where J_1 is the Bessel function of the first kind of order 1"""
    # TODO: Check if 0.5*x == 0 is ok - seems to be j1(x) < x
    # print 'jinc:', x
    is_zero=0.5*x==0
    result=j1(x)/np.where(is_zero, 1, x)
    return np.where(is_zero, 0.5, result)

def calc_kin_kout(h, qz_max, qz_step, a, lamda):
    """Calcualtes the k_in and k_out vectors for a reflectivity setup, assumin qy=0.

    Parameters:
        h(int): The inplane reflection index.
        qz_max (float): The maximum qz value wanted.
        qz_step (float): The step length in qz.
        a (float): The unit cell size in the x-direction.
        lamda (float): The wavelength of radiation used.

    Returns:
        k_in (array): A 3xM array representing the incident beam.
        k_out (array): A 3xM array represting the reflected beam.
    """
    qx=h*2*np.pi/a
    qz=np.arange(np.sqrt(2*2*np.pi/lamda*np.abs(qx)-qx**2)+qz_step, qz_max, qz_step)
    q_abs=np.sqrt(qx**2+qz**2)
    delta=np.arcsin(qx/q_abs)
    theta=np.arcsin(q_abs/4/np.pi*lamda)
    alpha=theta+delta
    beta=theta-delta
    k_in=2*np.pi/1.54*np.c_[np.cos(alpha), np.zeros_like(alpha), np.sin(alpha)].T
    k_out=2*np.pi/1.54*np.c_[np.cos(beta), np.zeros_like(beta), -np.sin(beta)].T
    return k_in, k_out

def ReflQ_ref(Q, lamda, n, d):
    """Parratt's recursion algorithm used as reference method in tests"""
    # Length of k-vector in vaccum
    d=d[1:-1]
    Q0=4*np.pi/lamda
    # Calculates the wavevector in each layer
    Qj=np.sqrt((n[:, np.newaxis]**2-n[-1]**2)*Q0**2+n[-1]**2*Q**2)
    # Fresnel reflectivity for the interfaces
    rp=(Qj[1:]-Qj[:-1])/(Qj[1:]+Qj[:-1])
    p=np.exp(1.0j*d[:, np.newaxis]*Qj[1:-1])  # Ignoring the top and bottom layer for the calc.
    # Setting up a matrix for the reduce function. Reduce only takes one array
    # as argument
    rpp=np.array(list(map(lambda x, y: [x, y], rp[1:], p)))

    # Paratt's recursion formula

    def formula(rtot, rint):
        return (rint[0]+rtot*rint[1])/(1+rtot*rint[0]*rint[1])

    # Implement the recursion formula
    r=reduce(formula, rpp, rp[0])
    # return the reflectivity
    return np.abs(r)**2

if __name__=='__main__':
    from pylab import *

    atype=np.complex128
    lamda=1.54
    if True:
        # Tests for the calculation of the fields at the vacuum interface.
        theta=arange(0, 5, 0.01)+1e-13
        Q=4*np.pi/lamda*np.sin(theta*np.pi/180)
        n_rep=1
        n=array([1-7.57e-6+1.73e-7j, ]+
                [1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j]*n_rep+
                [1.0, ], dtype=atype)
        d=array([2, ]+
                [80, 20]*n_rep+
                [2, ], dtype=atype)
        sigma=array([0, ]+
                    [0, 0]*n_rep+
                    [0])
        n_att=1
        import time

        t1=time.clock()
        for i in range(n_att):
            R_ref=ReflQ_ref(Q, 1.54, n, d)
        t2=time.clock()
        for i in range(n_att):
            E_T, E_R, k_j=calc_fields(Q/2.0, 1.54, (2*np.pi/1.54)**2*(1-n**2), d)
        t3=time.clock()
        iprint('Paratt: ', (t2-t1)/n_att)
        iprint('Abeles: ', (t3-t2)/n_att)
        R_try=np.abs(E_R[:, -1])**2
        semilogy(theta, R_ref)
        semilogy(theta, R_try, '.')

        legend(('Parratt', 'Abeles with roughness/interface layer'))
        xlabel('Two Theta [deg.]')
        ylabel('Reflectivity')
        show()

    if False:
        # Test case with a single rod
        a=2e4
        b=2e4

        v_lay=(2*pi/1.54)**2*(1-(1-(1.57-0.4J)*r_e*lamda**2/2/np.pi)**2)
        v_sub=(2*pi/1.54)**2*(1-(1-(0.71-0.09J)*r_e*lamda**2/2/np.pi)**2)
        # print v_lay
        v_lay=(2*pi/1.54)**2*(1-(1-7.57e-6+1.73e-7j)**2)
        # print v_lay
        d=array([0, 1, 0])
        vf_list=[mult(v_lay, ft_layer(a*b)), mult(v_lay, ft_circle(radius=a/4.)), mult(0, ft_layer(a*b))]
        vf_func=assemble_vf_func(vf_list)

        h=0
        qz_max=0.4
        qz_step=0.001

        k_in, k_out=calc_kin_kout(h, qz_max, qz_step, a, lamda)
        v_mean=vf_func([0, 0, 0])/a/b
        iprint(v_mean)
        vf=vf_func(k_out-k_in).T
        R=coherent_refl(k_in, k_out, lamda, vf, v_mean, d, a*b)
        R_kin=coherent_refl(k_in, k_out, lamda, vf, v_mean, d, a*b, kin=True)
        T_1, R_1, kz_1=calc_fields(k_in[2], lamda, v_mean, d)
        # T = coherent_nonspec_refl(k_in, k_out, lamda, vf, v_mean, d)
        # T_kin = coherent_nonspec_refl(k_in, k_out, lamda, vf, v_mean, d, kin=True)

        figure()
        qz=-(k_out[2]-k_in[2])
        semilogy(qz, np.abs(R))
        semilogy(qz, np.abs(R_1[:, -1])**2, '.')
        semilogy(qz, np.abs(R_kin))
        show()

    if False:
        # Do a h mapping
        a=2e4
        b=2e4

        v_lay=(2*pi/1.54)**2*(1-(1-(1.57-0.4J)*r_e*lamda**2/2/np.pi)**2)
        v_sub=(2*pi/1.54)**2*(1-(1-(0.71-0.09J)*r_e*lamda**2/2/np.pi)**2)
        # print v_lay
        # v_lay = (2*pi/1.54)**2*(1 - (1 - 7.57e-6 + 1.73e-7j)**2)
        # print v_lay
        d=array([2, 500, 50, 1])
        vf_list=[mult(v_lay, ft_layer(a*b)), mult(v_lay, ft_circle(radius=a/4.*1.5)),
                 mult(v_lay, ft_circle(radius=a/4.*1.5)), mult(0, ft_layer(a*b))]
        vf_func=assemble_vf_func(vf_list)

        h_max=6
        qz_max=0.4
        qz_step=0.001
        # k_in, k_out = calc_kin_kout(0, qz_max, qz_step, a, lamda)
        # print k_in
        # print k_out
        figure()
        # locator_params(axis='y', nbins=2)
        for h in range(0, h_max):
            k_in, k_out=calc_kin_kout(h, qz_max, qz_step, a, lamda)
            v_mean=vf_func([0, 0, 0])/a/b
            vf=vf_func(k_out-k_in).T
            R=coherent_refl(k_in, k_out, lamda, vf, v_mean, d, a*b)
            R_kin=coherent_refl(k_in, k_out, lamda, vf, v_mean, d, a*b, kin=True)
            qz=-(k_out[2]-k_in[2])
            ax=subplot(h_max, 1, h+1)
            plot(qz, R)
            plot(qz, R_kin, '--')
            yloc=plt.MaxNLocator(3)
            ax.yaxis.set_major_locator(yloc)

            # ax.locator_params(nbins=3, axis='y')
            ax.set_yscale('log')
            ax.set_xlim([0, qz_max])
            ax.set_ylim([1e-14, 10])
            ax.yaxis.set_ticks([1e-12, 1e-2])
            ax.text(0.95, 0.95, 'h = %d'%h,
                    verticalalignment='top',
                    horizontalalignment='right',
                    transform=ax.transAxes,
                    color='black', fontsize=15,
                    bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 3}
                    )
            if h!=h_max-1:
                ax.xaxis.set_ticklabels([])
        suptitle("Comparison between DWBA (full lines) and BA (dashed)")
        subplots_adjust(hspace=0)
        xlabel('Qz [1/AA]')
        show()



import warnings
from dataclasses import dataclass
from logging import debug

import matplotlib
import matplotlib.axes
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg
from matplotlib.figure import Figure

from numpy import *
import wx
from wx.py.editwindow import EditWindow
from ..core.config import BaseConfig, Configurable
from ..data import DataList, DataSet

# ==============================================================================
class SimplePlotPanel(wx.Panel):
    ax: matplotlib.axes.Axes

    def __init__(self, parent, id=-1, color=None, dpi=None,
                 style=wx.NO_FULL_REPAINT_ON_RESIZE | wx.EXPAND | wx.ALL, **kwargs):
        wx.Panel.__init__(self, parent, id=id, style=style, **kwargs)
        if dpi is None:
            dpi=wx.GetApp().dpi_scale_factor*96.  # wx.GetDisplayPPI()[0]
        self.parent=parent
        debug('init PlotPanel - setup figure')
        self.figure=Figure(figsize=(1.0, 1.0), dpi=dpi)
        debug('init PlotPanel - setup canvas')
        self.canvas=FigureCanvasWxAgg(self, -1, self.figure)
        self.SetColor(color)
        self._resizeflag=True
        self.print_size=(15./2.54, 12./2.54)
        # self._SetSize()

        # debug('init PlotPanel - bind events')
        self.Bind(wx.EVT_IDLE, self._onIdle)
        self.Bind(wx.EVT_SIZE, self._onSize)

        cursor=wx.Cursor(wx.CURSOR_CROSS)
        self.canvas.SetCursor(cursor)
        self.ax=self.figure.add_subplot(111)

        # Create the drawing bitmap
        self.bitmap=wx.Bitmap(1, 1, depth=wx.BITMAP_SCREEN_DEPTH)
        #        DEBUG_MSG("__init__() - bitmap w:%d h:%d" % (w,h), 2, self)
        self._isDrawn=False
        debug('end init PlotPanel')

    def clear(self):
        self.figure.clear()
        self.ax=self.figure.add_subplot(111)

    def SetColor(self, rgbtuple=None):
        ''' Set the figure and canvas color to be the same '''
        if not rgbtuple:
            rgbtuple=self.parent.GetBackgroundColour()
            # wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE).Get()
        col=[c/255. for c in rgbtuple]
        self.figure.set_facecolor(col)
        self.figure.set_edgecolor(col)
        self.canvas.SetBackgroundColour(wx.Colour(*rgbtuple))

    def _onSize(self, evt):
        self._resizeflag=True
        self._SetSize()

    def _onIdle(self, evt):
        if self._resizeflag:
            self._resizeflag=False
            self._SetSize()

    def _SetSize(self):
        ''' This method can be called to force the Plot to be a desired
            size which defaults to the ClientSize of the Panel.
        '''
        pixels=self.GetMinSize()

        self.canvas.SetSize(pixels)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            try:
                self.figure.tight_layout(h_pad=0)
            except ValueError:
                pass

    def flush_plot(self):
        # with warnings.catch_warnings():
        #     warnings.simplefilter("ignore", UserWarning)
        #     self.figure.tight_layout(h_pad=0)
        self.canvas.draw()

@dataclass
class PublicationConfig(BaseConfig):
    section = 'publication graph'

    width: float=3.39
    heigth: float=2.36
    font_size: int=int(matplotlib.rcParams['font.size'])
    font_familty: int=wx.FONTFAMILY_DEFAULT
    font_face: str=matplotlib.rcParams['font.sans-serif'][0]

    start_text: str=repr('''fig.set_facecolor('white')
# for other matplotlib rc style options see https://matplotlib.org/stable/tutorials/introductory/customizing.html
rcParams['font.family'] = 'sans-serif'
rcParams['font.size'] = font_size # dialog entry
rcParams['font.sans-serif'] = font_face # dialog entry
rcParams['font.weight'] = font_weight # dialog entry
rcParams['mathtext.fontset'] = 'dejavusans' # 'dejavuserif'|'cm'|'stix'|'stixsans'
clear()

for di in data:
    if not di.show:
        continue
    errorbar(di.x, di.y, yerr=di.error, label="data-"+di.name, **di.data_kwds)
    semilogy(di.x, di.y_sim, label="sim-"+di.name, **di.sim_kwds)
xlabel("q (Å$^{-1}$)")
ylabel("Reflectivity")
legend()
tight_layout(pad=0.5)
show()
#savefig(r"your_file_name.png", dpi=300)'''.splitlines())

class PublicationDialog(wx.Dialog, Configurable):
    opt: PublicationConfig

    def __init__(self, parent, id=-1, data: DataList=None):
        wx.Dialog.__init__(self, parent, id=id,
                           style=wx.MAXIMIZE_BOX|wx.MINIMIZE_BOX|wx.RESIZE_BORDER|wx.DEFAULT_DIALOG_STYLE)
        Configurable.__init__(self)
        self.ReadConfig()
        self.SetTitle('Custom plotting for Publication')

        self.data=data
        self.dpi_scale_factor=parent.dpi_scale_factor
        hbox=wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(hbox)
        vbox=wx.BoxSizer(wx.VERTICAL)
        hbox.Add(vbox, 10, wx.EXPAND|wx.ALL, 4)
        self.tinput=EditWindow(self, wx.ID_ANY)
        self.tinput.SetBackSpaceUnIndents(True)
        self.tinput.SetText('\n'.join(eval(self.opt.start_text)))

        vbox.Add(self.tinput, 1, wx.EXPAND, 0)
        button=wx.Button(self, label='plot')
        vbox.Add(button, 0, wx.EXPAND, 0)

        vbox2=wx.BoxSizer(wx.VERTICAL)
        hbox.Add(vbox2, 1, wx.EXPAND|wx.ALL, 4)
        self.plot=SimplePlotPanel(self)
        self.plot.SetMinSize(wx.Size(300,300))
        vbox2.Add(self.plot, 1, wx.EXPAND, 0)
        self.error_text=wx.StaticText(self)
        self.error_text.SetBackgroundColour(wx.Colour(255, 150, 150))
        self.error_text.Hide()
        vbox2.Add(self.error_text, 0, wx.EXPAND, 0)
        hbox2=wx.BoxSizer(wx.HORIZONTAL)
        vbox2.Add(hbox2, 0, wx.EXPAND, 0)
        self.width=wx.SpinCtrlDouble(self, min=0.1, max=30.0, initial=self.opt.width,
                                     inc=0.01, style=wx.SP_ARROW_KEYS)
        self.height=wx.SpinCtrlDouble(self, min=0.1, max=30.0, initial=self.opt.heigth,
                                      inc=0.01, style=wx.SP_ARROW_KEYS)
        hbox2.Add(wx.StaticText(self, label='Width (inches)'))
        hbox2.Add(self.width, 0, wx.FIXED_MINSIZE|wx.RIGHT, 10)
        hbox2.Add(wx.StaticText(self, label='Height (inches)'))
        hbox2.Add(self.height, 0, wx.FIXED_MINSIZE|wx.RIGHT, 10)
        hbox2=wx.BoxSizer(wx.HORIZONTAL)
        vbox2.Add(hbox2, 0, wx.EXPAND, 0)
        self.font=wx.FontPickerCtrl(self, style=wx.FNTP_FONTDESC_AS_LABEL)
        self.font.SetSelectedFont(wx.Font(self.opt.font_size, self.opt.font_familty,
                                          wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, False,
                                          faceName=self.opt.font_face))
        hbox2.Add(wx.StaticText(self, label='Font'))
        hbox2.Add(self.font, 0, wx.EXPAND, 0)

        hbox2=wx.BoxSizer(wx.HORIZONTAL)
        vbox2.Add(hbox2, 0, wx.EXPAND, 0)
        save_button=wx.Button(self, label='store current script and settings as default')
        hbox2.Add(save_button, 0, wx.EXPAND, 0)

        self.hbox=hbox
        self.Bind(wx.EVT_BUTTON, self.OnPlot, button)
        self.Bind(wx.EVT_BUTTON, self.OnStoreDefaults, save_button)
        self.Bind(wx.EVT_COMMAND_ENTER, self.OnPlot)
        self.Bind(wx.EVT_SPINCTRLDOUBLE, self.OnPlot)
        self.SetSize(wx.Size(800,800))

    def OnPlot(self, event):
        txt=self.tinput.GetText()
        font=self.font.GetSelectedFont()
        weights={wx.FONTWEIGHT_BOLD: 'bold', wx.FONTWEIGHT_NORMAL: 'normal'}
        env=dict(
            rcParams=matplotlib.rcParams,
            data=self.data,
            show=self.plot.flush_plot,
            fig=self.plot.figure,
            canvas=self.plot.canvas,
            savefig=self.plot.figure.savefig,
            tight_layout=self.plot.figure.tight_layout,

            font_size=font.GetPointSize(),
            font_face=font.GetFaceName(),
            font_weight=weights[font.GetWeight()],
            )
        def clear():
            self.plot.clear()
            ax=self.plot.ax
            env['plot']=ax.plot,
            env['semilogy']=ax.semilogy
            env['errorbar']=ax.errorbar
            env['title']=ax.set_title

            env['xlabel'] = ax.set_xlabel
            env['ylabel'] = ax.set_ylabel
            env['legend'] = ax.legend
            env['axes'] = ax
        env['clear']=clear

        w=int(self.width.GetValue()*72*self.dpi_scale_factor*2)
        h=int(self.height.GetValue()*72*self.dpi_scale_factor*2)
        self.plot.figure.dpi=int(72*self.dpi_scale_factor*2)
        size=wx.Size(w,h)
        self.plot.SetMinSize(size)
        self.plot.SetSize(size)
        clear()
        try:
            exec(txt, env)
        except Exception as e:
            self.error_text.SetLabel(e.__class__.__name__+': '+str(e))
            self.error_text.Show()
        else:
            self.error_text.Hide()

        # make sure we reset the default rc values
        matplotlib.rcdefaults()
        self.hbox.SetItemMinSize(self.plot, size)
        self.Layout()

    def OnStoreDefaults(self, event):
        self.opt.start_text=repr(self.tinput.GetText().splitlines())
        self.opt.width=self.width.GetValue()
        self.opt.heigth=self.height.GetValue()

        font=self.font.GetSelectedFont()
        self.opt.font_size=font.GetPointSize()
        self.opt.font_face=font.GetFaceName()
        self.opt.font_familty=font.GetFamily()
        self.WriteConfig(default=True)



from genx.core.custom_logging import iprint
from numpy import sqrt, pi, sin, cos, newaxis, ones, array, exp, r_, rollaxis, append, c_, arange, log10

def IntElfield(theta, lamda, n, d):
    # Length of k-vector in vaccum
    k=2*pi/lamda
    # Calculates the wavevector in each layer
    Qj=2*k*sqrt(n[:, newaxis]**2-cos(theta*pi/180)**2)
    # print Qj
    # Fresnel reflectivity for the interfaces
    rp=(Qj[1:]-Qj[:-1])/(Qj[1:]+Qj[:-1])
    tp=1+rp
    # tp=2*Qj[1:]/(Qj[1:]+Qj[:-1])
    # print tp.shape
    # Beta=k*d*n[:-1]*sin(theta*pi/180)
    # print Beta.shape
    D=d[:, newaxis]*ones(theta.shape)
    # E_p=[1]
    # E_m=[0]
    E_p=[1*ones(theta.shape)]
    E_m=[0*ones(theta.shape)]
    # print E_p[0].shape
    # print D[0].shape
    # print Qj[0].shape
    # for i in range(0,len(rp)):
    #    E_p.append(1/tp[i]*(E_p[i]*exp(-1.0J*D[i]*Qj[i]/2)+E_m[i]*rp[i]*exp(1.0J*D[i]*Qj[i]/2)))
    #    E_m.append(1/tp[i]*(E_m[i]*exp(1.0J*D[i]*Qj[i]/2)+E_p[i]*rp[i]*exp(-1.0J*D[i]*Qj[i]/2)))
    for i in range(0, len(rp)):
        E_p.append(1/tp[i]*(E_p[i]*exp(-1.0J*D[i]*Qj[i]/2)+E_m[i]*rp[i]*exp(1.0J*D[i]*Qj[i]/2)))
        E_m.append(1/tp[i]*(E_m[i]*exp(1.0J*D[i]*Qj[i]/2)+E_p[i]*rp[i]*exp(-1.0J*D[i]*Qj[i]/2)))
    # print E_p
    E_p=array(E_p[1:])
    E_m=array(E_m[1:])
    return E_p/E_p[-1], E_m/E_p[-1], Qj/2

def AmpElfield_q(k, kx, lamda, n, d):
    # Length of k-vector in vaccum
    k=2*pi/lamda
    # Calculates the wavevector in each layer
    # Qj=2*k*sqrt(n[:,newaxis]**2-cos(theta*pi/180)**2)
    Qj=2*sqrt(k**2*n[:, newaxis]**2-kx**2)
    # Fresnel reflectivity for the interfaces
    rp=(Qj[1:]-Qj[:-1])/(Qj[1:]+Qj[:-1])
    # print rp
    tp=1+rp
    # tp=2*Qj[1:]/(Qj[1:]+Qj[:-1])
    # print tp.shape
    # Beta=k*d*n[:-1]*sin(theta*pi/180)
    # print Beta.shape
    D=d[:, newaxis]*ones(kx.shape)
    # print D.shape
    # print rp.shape
    # E_p=[1]
    # E_m=[0]
    # old version
    E_p=[1*ones(kx.shape)]
    E_m=[0*ones(kx.shape)]
    for i in range(0, len(rp)):
        E_p.append(1/tp[i]*(E_p[i]*exp(-1.0J*D[i]*Qj[i]/2)+E_m[i]*rp[i]*exp(1.0J*D[i]*Qj[i]/2)))
        E_m.append(1/tp[i]*(E_m[i]*exp(1.0J*D[i]*Qj[i]/2)+E_p[i]*rp[i]*exp(-1.0J*D[i]*Qj[i]/2)))
    # Added 020905 for testing another T amplitude (The Transmitted amp. trough the interface)
    # print E_p
    # End old version
    E_p=array(E_p)
    E_m=array(E_m)
    # print E_p.shape
    # print Qj.shape
    dtemp=r_[d, 0]
    # print dtemp
    dtemp=dtemp[:, newaxis]*ones(kx.shape)
    E_p=E_p*exp(-1.0J*dtemp*Qj/2)
    E_m=E_m*exp(1.0J*dtemp*Qj/2)
    E_p=list(E_p)
    E_m=list(E_m)
    # End addition
    E_p.reverse()
    E_m.reverse()
    Qj=list(Qj)
    Qj.reverse()
    Qj=array(Qj)
    # Removed 020905 as said above and replaced
    E_p=array(E_p)  # Old version
    # E_p=array(E_p)
    # End addition
    E_m=array(E_m)
    q=Qj/2  # Old ver. replaced 020905 with:
    # return (E_p[:-1]/E_p[0],E_m[:-1]/E_p[0],q[:-1]) # Last correspond to kz
    # testar amplitud i lager j+1 istallet
    return E_p[1:]/E_p[0], E_m[1:]/E_p[0], q[1:]  # Last correspond to kz

def AmpElfield2(kx, k, n, z):
    kz=sqrt(n[:, newaxis]**2*k**2-kx**2)
    r=(kz[:-1]-kz[1:])/(kz[:-1]+kz[1:])
    t=1+r
    X=0*ones(kx.shape)
    for i in range(len(n)-2, -1, -1):
        X=exp(-2j*kz[i]*z[i])*(r[i]+X*exp(2j*kz[i+1]*z[i]))/(1+r[i]*X*exp(2j*kz[i+1]*z[i]))
        iprint(i)
    # X=reflected amplitude...
    r=(kz[1:]-kz[:-1])/(kz[:-1]+kz[1:])
    t=1+r
    R=[X]
    T=[1*ones(X.shape)]
    for i in range(0, len(n)-1, 1):
        R.append(1/t[i]*(T[i]*r[i]*exp(-1.0j*(kz[i+1]+kz[i])*z[i])+R[i]*exp(-1.0j*(kz[i+1]-kz[i])*z[i])))
        T.append(1/t[i]*(T[i]*exp(1.0j*(kz[i+1]-kz[i])*z[i])+R[i]*r[i]*exp(1.0j*(kz[i+1]+kz[i])*z[i])))
    # R[len(n)-1]=0*ones(X.shape)
    R=array(R)
    T=array(T)
    return T, R, kz

from .xrmr import dot2, inv2

def AmpElfield_test(th, lam, n, d, dz=0.1):
    ''' Added 20110317 to see if I can do calcs of the fields inside a multilayer '''
    kz=2*pi/lam*sin(th*pi/180)
    k=2*pi/lam
    # n = r_[n, n[-1]]
    # d = r_[d, d[-1]]
    kz=sqrt((n[:, newaxis]**2-1)*k**2+kz**2)
    # print kz.shape, d.shape
    a=exp(-1.0J*kz*d[:, newaxis])[1:, :]
    # print a.shape
    # print kz[newaxis, newaxis, :, :].shape
    r=(kz[:-1]-kz[1:])/(kz[:-1]+kz[1:])
    M=1/kz[newaxis, newaxis, :-1, :]/2*array([[(kz[1:]+kz[:-1])*a, (kz[:-1]-kz[1:])/a],
                                              [(kz[:-1]-kz[1:])*a, (kz[1:]+kz[:-1])/a]])
    # M = array([[a, r*a],[r/a, 1/a]])
    M=rollaxis(M, 2, 0)[::, :, :, :]
    # "quick" version
    # MM = reduce(dot2, M)
    MMcum=[M[0]]
    for i in range(1, M.shape[0]):
        MMcum.append(dot2(MMcum[-1], M[i]))
    MM=MMcum[-1]

    R0=MM[1, 0]/MM[0, 0]
    TN=1/M[0, 0]
    R=abs(R0)**2
    T=abs(TN)**2

    # Calc el. fileds as a function of z
    Rlay=R0
    Tlay=1.0
    z=array([0])
    iprint('R0 ', R0.shape)
    E=array(abs(R0+1.0))[:, newaxis]*ones((th.shape[0], 1))
    # E = array([])
    iprint(kz.shape)
    iprint((th.shape[0], 1))
    iprint(ones((th.shape[0], 1)).shape)
    iprint(E.shape)
    for di, kzi, M, in zip(d[1:], kz[1:], MMcum):
        Mlay=inv2(M)
        Rlay_new=Mlay[0, 0]*Tlay+Mlay[0, 1]*Rlay
        Tlay_new=Mlay[1, 0]*Tlay+Mlay[1, 1]*Rlay
        znew=arange(0, di, dz)
        z=append(z, znew+z[-1])
        # print kzi.shape
        a=exp(-1.0J*kzi[:, newaxis]*(di-znew))
        # print (abs(Rlay_new[:, newaxis]*a + Tlay_new[:, newaxis]/a)**2).shape
        E=c_[E, abs(Rlay_new[:, newaxis]*a+Tlay_new[:, newaxis]/a)]
        Rlay=Rlay_new
        Tlay=Tlay_new

    # print MM.shape
    iprint(E.shape)
    # E = E.reshape((th.shape[0], z.shape[0]))
    return T, R, z, E

if __name__=='__main__':
    # R=[]
    # for x in arange(0,2,0.01):
    #    (E_p,E_m)=IntElfield(x,1.54,array([1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1]),array([0,80,20,80,20,80,20]))
    #    R.append(abs(E_m[-1]/E_p[-1])**2)
    import pylab

    theta=arange(0, 1+0.001666670, 0.005)
    # (E_p,E_m,Qj)=IntElfield(theta,1.54,array([1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1]),array([0,80,20,80,20,80,20]))
    # (E_p,E_m,Qj)=IntElfield(theta,1.540562,array([1-.15E-04/2+0.35E-06j/2,1-.45E-04/2+0.60E-05j/2,1]),array([0,1000]))
    from .paratt import Refl

    # c=Refl(theta,1.54,array([1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1-2.24e-5+2.89e-6j,1-7.57e-6+1.73e-7j,1]),array([80,20,80,20,80,20]),0)
    c=Refl(theta, 1.54, array([1-7.57e-6+1.73e-7j, 1-2.24e-5+2.89e-6j, 1]), array([0, 1000, 0]), array([0, 0, 0]))
    # gplt.plot(theta,log10(abs(E_m[-1]/E_p[-1])**2),theta,log10(abs(c**2)))
    # with open('Specrefl.dat','r') as f:
    #     t=io.read_array(f)
    #     t=transpose(t)
    # gplt.plot(theta,log10(abs(E_m[-1]/E_p[-1])**2),t[0],log10(t[1]))
    k=2*pi/1.54
    n=[1, 1-2.24e-5+2.89e-6j, 1-7.57e-6+1.73e-7j]
    n=array(n)
    z=array([0, -1000])
    # (T,R,k)=AmpElfield2(k*cos(theta*pi/180),k*ones(theta.shape), n,z)
    # (T,R,k)=IntElfield(theta,1.54, n, array([0,1000]))
    (T, R, zc, E)=AmpElfield_test(theta, 1.54, n, array([0, 1000, 0]))
    # print T[0]
    pylab.subplot(211)
    pylab.plot(theta, log10(abs(R)), theta, log10(abs(c)), '.-')
    pylab.ylabel('R')
    pylab.xlabel('theta [deg]')
    pylab.legend(('Abeles', 'Parratt'))
    pylab.subplot(212)
    pylab.contourf(theta, zc, E.transpose())
    # print zc.shape, zc[-1]
    # print log10(E[0,:]).shape
    # pylab.plot(zc, (E[1,:]))
    pylab.show()



'''
Controller class for the differnetial evolution class diffev
Takes care of stopping and starting - output to the gui as well
as some input from dialog boxes.
'''
import numpy as np
import time
from typing import Union
from threading import Thread, Event

import wx
import wx.lib.newevent

from .exception_handling import CatchModelError
from .settings_dialog import SettingsDialog
from .. import diffev, fom_funcs, model_control, levenberg_marquardt
from ..core.custom_logging import iprint
from ..solver_basis import SolverParameterInfo, SolverResultInfo, SolverUpdateInfo, GenxOptimizerCallback


# Custom events needed for updating and message parsing between the different
# modules.
(update_plot, EVT_UPDATE_PLOT)=wx.lib.newevent.NewEvent()
(update_text, EVT_SOLVER_UPDATE_TEXT)=wx.lib.newevent.NewEvent()
(update_parameters, EVT_UPDATE_PARAMETERS)=wx.lib.newevent.NewEvent()
(fitting_ended, EVT_FITTING_ENDED)=wx.lib.newevent.NewEvent()
(autosave, EVT_AUTOSAVE)=wx.lib.newevent.NewEvent()

class GuiCallbacks(GenxOptimizerCallback):
    def __init__(self, parent: wx.Window):
        self.parent=parent

    def text_output(self, text):
        '''
        Function to present the output from the optimizer to the user.
        Takes a string as input.
        '''
        evt=update_text(text=text)
        wx.QueueEvent(self.parent, evt)

    def plot_output(self, update_data):
        '''
        Solver to present the graphical output from the optimizer to the
        user. Takes the solver as input argument and picks out the
        variables to show in the GUI.
        '''
        evt=update_plot(data=update_data.data, fom_value=update_data.fom_value,
                        fom_name=update_data.fom_name,
                        fom_log=update_data.fom_log, update_fit=update_data.new_best,
                        desc='Fitting update')
        wx.QueueEvent(self.parent, evt)

    def parameter_output(self, param_info):
        '''
        Function to send an update event to update windows that displays
        the parameters to update the values.
        Takes the solver as input argument and picks out the variables to
        show in the GUI.
        '''
        evt=update_parameters(values=param_info.values,
                              new_best=param_info.new_best,
                              population=param_info.population,
                              max_val=param_info.max_val,
                              min_val=param_info.min_val,
                              fitting=True,
                              desc='Parameter Update', update_errors=False,
                              permanent_change=False)
        wx.QueueEvent(self.parent, evt)

    def fitting_ended(self, result_data):
        '''
        function used to post an event when the fitting has ended.
        This must be done since it is not thread safe otherwise. Same GUI in
        two threads when dialogs are run. dangerous...
        '''
        evt=fitting_ended(start_guess=result_data.start_guess,
                          error_message=result_data.error_message,
                          values=result_data.values,
                          new_best=result_data.new_best,
                          population=result_data.population,
                          max_val=result_data.max_val,
                          min_val=result_data.min_val,
                          fitting=True, desc='Fitting Ended')
        wx.QueueEvent(self.parent, evt)

    def autosave(self):
        '''
        Function that conducts an autosave of the model.
        '''
        evt=autosave()
        wx.QueueEvent(self.parent, evt)

class DelayedCallbacks(Thread, GuiCallbacks):
    last_text: Union[str, None]=None
    last_param: Union[SolverParameterInfo, None]=None
    last_update: Union[SolverUpdateInfo, None]=None
    last_endet: Union[SolverResultInfo, None]=None
    min_time=0.5
    last_iter: float=0.0
    wait_lock: Event
    stop_thread: Event

    def __init__(self, parent: wx.Window):
        GuiCallbacks.__init__(self, parent)
        Thread.__init__(self, daemon=True, name="GenxDelayedCallbacks")
        self.wait_lock=Event()
        self.stop_thread=Event()

    def run(self):
        self.last_iter=time.time()
        self.stop_thread.clear()
        while not self.stop_thread.is_set():
            # main loop for checking for updates and sending GUI events
            time.sleep(max(0., (self.last_iter-time.time()+self.min_time)))
            if self.last_text:
                GuiCallbacks.text_output(self, self.last_text)
                self.last_text=None
            if self.last_param:
                GuiCallbacks.parameter_output(self, self.last_param)
                self.last_param=None
            if self.last_update:
                GuiCallbacks.plot_output(self, self.last_update)
                self.last_update=None
            if self.last_endet:
                GuiCallbacks.fitting_ended(self, self.last_endet)
                self.last_endet=None
            self.last_iter=time.time()
            self.wait_lock.clear()
            self.wait_lock.wait()

    def exit(self):
        self.stop_thread.set()
        self.wait_lock.set()
        self.join(timeout=1.0)

    def text_output(self, text):
        self.last_text=text
        self.wait_lock.set()

    def fitting_ended(self, result_data):
        self.last_endet=result_data
        self.wait_lock.set()

    def parameter_output(self, param_info):
        self.last_param=param_info
        self.wait_lock.set()

    def plot_output(self, update_data):
        self.last_update=update_data
        self.wait_lock.set()

class ModelControlGUI:
    '''
    Class to take care of the GUI - solver interaction.
    Implements dialogboxes for setting parameters and controls
    for the solver routine. This is where the application specific
    code are used i.e. interfacing the optimization code to the GUI.
    '''

    def __init__(self, parent):
        self.parent=parent
        self.solvers={
            'Differential Evolution': diffev.DiffEv(),
            'Levenberg-Marquardt': levenberg_marquardt.LMOptimizer()
            }
        try:
            from ..bumps_optimizer import BumpsOptimizer
        except ImportError:
            pass
        else:
            self.solvers['Bumps']=BumpsOptimizer()

        self.controller=model_control.ModelController(self.solvers['Differential Evolution'])
        self.callback_controller=DelayedCallbacks(parent)
        self.callback_controller.start()
        self.controller.set_callbacks(self.callback_controller)
        self.parent.Bind(EVT_FITTING_ENDED, self.OnFittingEnded)
        self.parent.Bind(EVT_AUTOSAVE, self.AutoSave)

        # Now load the default configuration
        self.ReadConfig()

    def get_solvers(self):
        return list(self.solvers.keys())

    def set_solver(self, name):
        self.controller.optimizer=self.solvers[name]
        self.controller.set_callbacks(self.callback_controller)

    def ReadConfig(self):
        '''
        Reads the parameter that should be read from the config file.
        And set the parameters in both the optimizer and this class.
        '''
        self.controller.ReadConfig()

    def WriteConfig(self):
        '''
        Writes the current configuration of the solver to file.
        '''
        self.controller.WriteConfig()

    def ParametersDialog(self, frame):
        '''
        Shows the Parameters dialog box to set the parameters for the solver.
        '''
        # Update the configuration if a model has been loaded after
        # the object have been created..
        self.ReadConfig()
        fom_func_name=self.controller.model.fom_func.__name__
        if not fom_func_name in fom_funcs.func_names:
            ShowWarningDialog(self.parent, 'The loaded fom function, ' \
                              +fom_func_name+', does not exist '+ \
                              'in the local fom_funcs file. The fom fucntion has been'+
                              ' temporary added to the list of availabe fom functions')
            fom_funcs.func_names.append(fom_func_name)
            exectext='fom_funcs.'+fom_func_name+ \
                     ' = self.parent.model.fom_func'
            exec(exectext, locals(), globals())

        combined_options=self.controller.model.solver_parameters|self.controller.optimizer.opt
        dlg=SettingsDialog(frame, combined_options, title='Optimizer Settings')

        res=dlg.ShowModal()
        if res==wx.ID_OK:
            self.controller.model.WriteConfig()
            self.controller.optimizer.WriteConfig()
        dlg.Destroy()

    def ModelLoaded(self):
        '''
        Function that takes care of resetting everything when a model has
        been loaded.
        '''
        evt=update_plot(model=self.controller.get_fitted_model(),
                        fom_log=self.controller.get_fom_log(), update_fit=False,
                        desc='Model loaded')
        wx.PostEvent(self.parent, evt)

        # Update the parameter plot ...
        if self.controller.is_configured():
            # remember to add a check
            res=self.controller.get_result_info()
            try:
                evt=update_parameters(values=res.values,
                                      new_best=False,
                                      population=res.population,
                                      max_val=res.par_max,
                                      min_val=res.par_min,
                                      fitting=True,
                                      desc='Parameter Update', update_errors=False,
                                      permanent_change=False)
            except AttributeError:
                iprint('Could not create data for parameters')
            else:
                wx.PostEvent(self.parent, evt)

    def OnFittingEnded(self, evt):
        '''
        Callback when fitting has ended. Takes care of cleaning up after
        the fit. Calculates errors on the parameters and updates the grid.
        '''
        if evt.error_message:
            ShowErrorDialog(self.parent, evt.error_message)
            return

        message='Do you want to keep the parameter values from the fit?'
        dlg=wx.MessageDialog(self.parent, message, 'Keep the fit?', wx.YES_NO | wx.ICON_QUESTION)
        if dlg.ShowModal()==wx.ID_YES:
            evt = update_parameters(values=evt.values,
                                    new_best=True,
                                    population=evt.population,
                                    max_val=evt.max_val,
                                    min_val=evt.min_val,
                                    fitting=False,
                                    desc='Parameter Update', update_errors=False,
                                    permanent_change=True)
            wx.PostEvent(self.parent, evt)
        else:
            evt = update_parameters(values=evt.start_guess,
                                    new_best=True,
                                    population=evt.population,
                                    max_val=evt.max_val,
                                    min_val=evt.min_val,
                                    fitting=False,
                                    desc='Parameter Update', update_errors=False,
                                    permanent_change=False)
            wx.PostEvent(self.parent, evt)

    def CalcErrorBars(self):
        return self.controller.CalcErrorBars()

    def ProjectEvals(self, parameter):
        return self.controller.ProjectEvals(parameter)

    def ScanParameter(self, parameter, points):
        '''
        Scans one parameter and records its fom value as a function 
        of the parameter value.
        '''
        row=parameter
        model=self.controller.model
        (funcs, vals)=model.get_sim_pars()
        minval=model.parameters.get_data()[row][3]
        maxval=model.parameters.get_data()[row][4]
        parfunc=funcs[model.parameters.get_sim_pos_from_row(row)]
        par_def_val=vals[model.parameters.get_sim_pos_from_row(row)]
        step=(maxval-minval)/points
        par_vals=np.arange(minval, maxval+step, step)
        fom_vals=np.array([])

        par_name=model.parameters.get_data()[row][0]
        dlg=wx.ProgressDialog("Scan Parameter",
                              "Scanning parameter "+par_name,
                              maximum=len(par_vals),
                              parent=self.parent,
                              style=wx.PD_APP_MODAL | wx.PD_ELAPSED_TIME
                                    | wx.PD_REMAINING_TIME | wx.PD_AUTO_HIDE)
        with CatchModelError(self.parent, 'ScanParameter', 'scan through values') as cme:
            # Start with setting all values
            [f(v) for (f, v) in zip(funcs, vals)]
            for par_val in par_vals:
                parfunc(par_val)
                fom_vals=np.append(fom_vals, model.evaluate_fit_func())
                dlg.Update(len(fom_vals))
        dlg.Destroy()
        # resetting the scanned parameter
        parfunc(par_def_val)
        if cme.successful:
            return par_vals, fom_vals

    def ResetOptimizer(self):
        pass

    def StartFit(self):
        self.controller.StartFit()

    def StopFit(self):
        self.controller.StopFit()

    def ResumeFit(self):
        self.controller.ResumeFit()

    def IsFitted(self):
        return self.controller.IsFitted()

    def AutoSave(self, _event):
        self.controller.save()

    def load_file(self, fname):
        self.controller.load_file(fname)
        solver_classes=[si.__class__ for si in self.solvers.values()]
        loaded_solver=self.controller.optimizer.__class__
        if loaded_solver in solver_classes:
            current_solver=list(self.solvers.keys())[solver_classes.index(loaded_solver)]
        else:
            self.solvers[loaded_solver.__name__]=self.controller.optimizer
            current_solver=loaded_solver.__name__
            self.parent.eh_ex_add_solver_selection(current_solver)
        self.parent.eh_ex_set_solver_selection(current_solver)

    def set_error_bars_level(self, value):
        '''
        Sets the value of increase of the fom used for errorbar calculations
        '''
        if value<1:
            raise ValueError('fom_error_bars_level has to be above 1')
        else:
            self.controller.optimizer.opt.errorbar_level=value

    def set_save_all_evals(self, value):
        '''
        Sets the boolean value to save all evals to file
        '''
        self.controller.optimizer.opt.save_all_evals=bool(value)


def ShowWarningDialog(frame, message):
    dlg=wx.MessageDialog(frame, message,
                         'Warning',
                         wx.OK | wx.ICON_WARNING
                         )
    dlg.ShowModal()
    dlg.Destroy()

def ShowErrorDialog(frame, message, position=''):
    if position!='':
        dlg=wx.MessageDialog(frame, message+'\n'+'Position: '+position,
                             'ERROR',
                             wx.OK | wx.ICON_ERROR
                             )
    else:
        dlg=wx.MessageDialog(frame, message,
                             'ERROR',
                             wx.OK | wx.ICON_ERROR
                             )
    dlg.ShowModal()
    dlg.Destroy()




"""
GenX model and optimizer control classes. All functional aspects should be covered here (no GUI).
"""
import os
import sys
import h5py
from logging import warning

from .core.config import config
from .exceptions import ErrorBarError, GenxIOError
from .model import Model
from .solver_basis import GenxOptimizer, GenxOptimizerCallback

class ModelController:
    def __init__(self, optimizer: GenxOptimizer):
        self.model=Model()
        self.optimizer=optimizer

    def set_callbacks(self, callbacks: GenxOptimizerCallback):
        self.optimizer.set_callbacks(callbacks)

    def is_configured(self):
        return self.optimizer.is_configured()

    def get_result_info(self):
        return self.optimizer.get_result_info()

    def get_fitted_model(self):
        return self.optimizer.get_model()

    def get_fom_log(self):
        return self.optimizer.get_fom_log()

    def ReadConfig(self):
        '''
        Reads the parameter that should be read from the config file.
        And set the parameters in both the optimizer and this class.
        '''
        self.model.ReadConfig()
        self.optimizer.ReadConfig()

    def WriteConfig(self):
        '''
        Writes the current configuration of the solver to file.
        '''
        self.model.WriteConfig()
        self.optimizer.WriteConfig()

    def CalcErrorBars(self):
        '''
        Method that calculates the errorbars for the fit that has been
        done. Note that the fit has to been conducted before this is run.
        '''
        if self.optimizer.n_fom_evals==0:
            raise ErrorBarError('Can not find any stored evaluations of the model in the optimizer.\n'
                                'Run a fit before calculating the errorbars.')
        if self.optimizer.get_start_guess() is not None and not self.optimizer.is_running():
            n_elements=len(self.optimizer.get_start_guess())
            error_values=[]
            for index in range(n_elements):
                # calculate the error, this is threshold based and not rigours
                (error_low, error_high)=self.optimizer.calc_error_bar(index)
                error_str='(%.3e, %.3e)'%(error_low, error_high)
                error_values.append(error_str)
            return error_values
        else:
            raise ErrorBarError('Wait for fit to finish or fit to changed model first.')

    def ProjectEvals(self, parameter: int):
        '''
        Projects the parameter number parameter on one axis and returns
        the fom values.
        '''
        row=self.model.parameters.get_pos_from_row(parameter)
        if self.optimizer.get_start_guess() is not None and not self.optimizer.is_running():
            return self.optimizer.project_evals(row)
        else:
            raise ErrorBarError()

    def StartFit(self):
        '''
        Function to start running the fit
        '''
        # Make sure that the config of the solver is updated..
        self.optimizer.ReadConfig()
        # Reset all the errorbars
        self.model.parameters.clear_error_pars()
        self.optimizer.start_fit(self.model)

    def StopFit(self):
        '''
        Function to stop a running fit
        '''
        self.optimizer.stop_fit()

    def ResumeFit(self):
        '''
        Function to resume the fitting after it has been stopped
        '''
        self.optimizer.ReadConfig()
        self.optimizer.resume_fit(self.model)

    def IsFitted(self):
        '''
        Returns true if a fit has been started otherwise False
        '''
        return self.optimizer.is_fitted()

    def save(self):
        self.save_file(self.model.get_filename())

    def save_file(self, fname: str):
        """
        Saves objects model, optimiser and config into file fname
        """
        if fname.endswith('.gx'):
            self.save_gx(fname)
        elif fname.endswith('.hgx'):
            self.save_hgx(fname)
        else:
            raise GenxIOError('Wrong file ending, should be .gx or .hgx')
        self.model.filename=os.path.abspath(fname)
        self.model.saved=True

    def load_file(self, fname: str):
        """
        Loads parameters from fname into model, optimizer and config
        """
        if fname.endswith('.gx'):
            self.load_gx(fname)
        elif fname.endswith('.hgx'):
            self.load_hgx(fname)
        else:
            raise GenxIOError('Wrong file ending, should be .gx or .hgx')
        self.model.filename=os.path.abspath(fname)

    def save_hgx(self, fname: str):
        f=h5py.File(fname.encode('utf-8'), 'w')
        g=f.create_group(self.model.h5group_name)
        self.model.write_h5group(g)
        opt_group=g.create_group(self.optimizer.h5group_name)
        opt_group['solver']=self.optimizer.__class__.__name__
        opt_group['solver_module']=self.optimizer.__class__.__module__
        self.optimizer.write_h5group(opt_group)
        self.WriteConfig()
        g['config']=config.model_dump().encode('utf-8')
        f.close()

    def load_hgx(self, fname: str):
        f=h5py.File(fname.encode('utf-8'), 'r')
        g=f[self.model.h5group_name]
        self.model.read_h5group(g)
        opt_group=g[self.optimizer.h5group_name]
        try:
            solver_class=opt_group.get('solver')[()]
            solver_module=opt_group.get('solver_module')[()]
        except (KeyError, TypeError):
            solver_class='DiffEv'
            solver_module='genx.diffev'
        else:
            if type(solver_class) is bytes:
                solver_class = solver_class.decode('utf-8')
                solver_module = solver_module.decode('utf-8')
        if solver_class!=self.optimizer.__class__.__name__:
            try:
                exec(f'from {solver_module} import {solver_class}')
            except ImportError:
                warning(f'Could not import solver {solver_class} from moudle {solver_module}')
            else:
                prev_optimizer=self.optimizer
                exec(f'self.optimizer={solver_class}()')
                self.optimizer.set_callbacks(prev_optimizer.get_callbacks())
        self.optimizer.read_h5group(opt_group)
        try:
            config.load_string(g['config'][()].decode('utf-8'))
            self.ReadConfig()
        except KeyError:
            pass
        except AttributeError:
            pass
        f.close()

    def save_gx(self, fname: str):
        self.model.save(fname)
        self.model.save_addition('config', config.model_dump())
        self.model.save_addition('optimizer', self.optimizer.pickle_string(clear_evals=
                                        not config.getboolean('solver', 'save all evals')))

    def load_gx(self, fname: str):
        self._patch_modules() # for compatibility with old files
        self.model.load(fname)
        config.load_string(self.model.load_addition('config').decode('utf-8'))
        self.optimizer.pickle_load(self.model.load_addition('optimizer'))

    def _patch_modules(self):
        # add legacy items to genx for loading of pickled strings from old program
        from genx import diffev
        diffev.defualt_autosave=diffev.DiffEv._callbacks.autosave


'''Utilities for GenX and scattering
=================================

This library contains nice to have functions and classes. Most
noteworthy is the UserVars class. The scattering factor tables can also
come handy.

Classes
-------

UserVars
~~~~~~~~
This class is used to contain user defined variables which is problem
specific and do not warrant a completly new model library. It is
contstructed as: ``vars = UserVars()``.

A new variable is created by calling the new_var method:
``vars.new_var('my_var', 10.0)``

The methods in this class are:

``__init__(self)``
   Create a new variable container
``new_var(self, name, value)``
   Create a new varaible with name [string] name and value value [float]

The created varaibles are accesed through the as [Uservars].set[Name]
and [Uservars].get[Name]. In the example above this would read:

``vars.setMy_var(20)`` to set a value and ``vars.getMy_vars()`` to
retrieve the value.

Scattering factor databases
~~~~~~~~~~~~~~~~~~~~~~~~~~~
The following databases exist in GenX. All databases are accesed through
``[name].[element]`` for example: ``fp.fe``

fp
^^
The fp contains the scattering factors at Q = 0 for the 92 first
elements up to U in units of electrons or more precise Thompson
scattering lengths. The data is taken from
http://henke.lbl.gov/optical_constants/asf.html. These tables are also
known as the henke tables. According to cxros homepage: There are 500+
points on a uniform logarithmic mesh from 10 to 30,000 eV with points
added 0.1 eV above and below "sharp" absorption edges. For some elements
data on a finer mesh includes structure around absorption edges. (Below
29 eV f1 is set equal to -9999.)

bc
^^
The bc data base contains the coherent scattering length for neutrons
according to the data published in Neutron News, Vol. 3, No. 3, 1992,
pp. 29-37. The data file is taken from the Dabax library compiled by
esrf http://ftp.esrf.eu/pub/scisoft/xop2.3/DabaxFiles/. To use isotopes
just add "i" in front of the name i.e., bc.i57Fe.

f
^
The f database contains the *scattering vector dependent* isotropic
scattering factors in electrons or more precise Thompson scattering
lengths. The data is taken from the fp and the f0 database.

f0
^^
The f0 database contains the *scattering vector dependent* isotropic
scattering factors in electrons or more precise Thompson scattering
lengths. The data is the so-called Croemer-Mann tables see:
International Tables vol. 4 or vol C; in vol. C refer to pg 500-502.
Note, this is only good out to sin(theta)/lambda < 2.0 [Angstrom^-1].
The data is also fetched from the Dabax library at:
http://ftp.esrf.eu/pub/scisoft/xop2.3/DabaxFiles/ This database is the
non-dispersive (without resonant contribution).

fw
^^
Same thing as f but scaled so that is can be used with a density in
g/cm\ :sup:`3\ .`

bw
^^
Same thing as bc but scaled so that is can be used with a density in
g/cm\ :sup:`3\ .`
'''
from .lib import scatteringlengths as sl
import os
from genx.core.custom_logging import iprint

_head, _tail=os.path.split(__file__)
# Look only after the file name and not the ending since
# the file ending can be pyc if compiled... 
__FILENAME__=_tail.split('.')[0]
# This assumes that plugin is under the current dir may need 
# changing
__MODULE_DIR__=_head
if __MODULE_DIR__=='':
    __MODULE_DIR__='.'

class UserVars:
    def __init__(self):
        self._penalty_funcs=[]

    def newVar(self, name, value):
        # Adds a new user variable to the class
        setattr(self, name, value)
        setattr(self, 'set'+name[0].upper()+name[1:], lambda value: setattr(self, name, value))
        setattr(self, 'get'+name[0].upper()+name[1:], lambda: getattr(self, name, value))

    new_var=newVar

    def new_sys_err(self, name, value, error, weight=1.0, correction=0.0):
        # Adds a new systematic error variable to the class that biases the FOM when changed from start value
        setattr(self, name, value)
        setattr(self, 'set'+name[0].upper()+name[1:], lambda v: setattr(self, name, v))
        setattr(self, 'get'+name[0].upper()+name[1:], lambda: getattr(self, name, value))
        setattr(self, 'penalty'+name[0].upper()+name[1:],
                lambda: weight*(((value-getattr(self, name, value))/error)**2-correction))
        self._penalty_funcs.append(getattr(self, 'penalty'+name[0].upper()+name[1:]))

# ==============================================================================
# Now create default databases for scattering lengths and form factors
# for now only x-rays at single wavelengths
# The non-disersive part but angle dependent
__f0_dict__, __rho0_dict__=sl.load_f0dabax(__MODULE_DIR__+
                                           '/databases/f0_CromerMann.dat',
                                           create_rho=True)
# Workaround for the oxygen
__f0_dict__['o2m']=__f0_dict__['o2-.']
__rho0_dict__['o2m']=__rho0_dict__['o2-.']
f0=sl.ScatteringLength(__f0_dict__)
# Dispersive part at Q = 0
__lookup_fp__=sl.create_fp_lookup(__MODULE_DIR__+'/databases/f1f2_nist/')
__lookup_fp_old__=sl.create_fp_lookup(__MODULE_DIR__+'/databases/f1f2_cxro/')

def create_fp(wavelength):
    return sl.FormFactor(wavelength, __lookup_fp__)

def create_fp_old(wavelength):
    return sl.FormFactor(wavelength, __lookup_fp_old__)

fp=create_fp(1.54)
fp_old=create_fp_old(1.54)
# The total angle dependent form factor
__lookup_f__=sl.create_f_lookup(__lookup_fp__, __f0_dict__)
f=sl.FormFactor(1.54, __lookup_f__)
# The electrondensity of an atom
__lookup_rho__=sl.create_rho_lookup(__lookup_fp__, __rho0_dict__,
                                    __f0_dict__)
rho=sl.FormFactor(1.54, __lookup_rho__)
# The coherent scattering length for neutrons
__bc_dict__=sl.load_bdabax(__MODULE_DIR__+'/databases/DeBe_NeutronNews.dat')
bc=sl.ScatteringLength(__bc_dict__)
# print 'Loading atomic weights'
__w_dict__=sl.load_atomic_weights_dabax(__MODULE_DIR__+ \
                                        '/databases/AtomicWeights.dat')
# print 'Making bw dict'
__bw_dict__=sl.create_scatt_weight(__bc_dict__, __w_dict__)
# print 'Making bw scattering lengths'
bw=sl.ScatteringLength(__bw_dict__)
# print 'Making fw scattering lengths'
__lookup_fw__=sl.create_fw_lookup(__lookup_fp__, __w_dict__)

def create_fw(wavelength):
    return sl.FormFactor(wavelength, __lookup_fw__)

fw=create_fw(1.54)

__f_chantler_dict__=sl.read_dabax(__MODULE_DIR__+'/databases/f1f2_Chantler.dat')
__lookup_fpc__=sl.load_fdabax(__MODULE_DIR__+'/databases/f1f2_Chantler.dat')
fpc=sl.FormFactor(15.4, __lookup_fpc__)

__lookup_fdisp__=sl.create_fpdisp_lookup(__MODULE_DIR__+'/databases/f1f2_nist/')
fd=sl.Database()
object.__setattr__(fd, 'lookup_value', __lookup_fdisp__)

if __name__=='__main__':
    # MyVars=UserVars()
    # MyVars.newVar('a',3)
    # print 'f0.fe(0)', f0.fe(0)
    # fp.set_wavelength(1.54)
    # print 'fp.fe', fp.fe
    # f.set_wavelength(1.54)
    # print 'f.fe(0)', f.fe(0)
    # print 'f.fe2p(0)', f.fe2p(0)
    # fe = np.array(__f_chantler_dict__['fe'])
    # print fe.reshape(len(fe)/7, 7)
    ##print bc.fe
    # print fpc.Sn
    # print fp.Sn
    import time

    N=1000
    t1=time.time()
    for i in range(N):
        fd.Fe(100.0)
    t2=time.time()
    iprint("Dispersive database access time: ", (t2-t1)/N)
    t1=time.time()
    for i in range(N):
        ignore=fp.Fe
    t2=time.time()
    iprint("Normal database access time: ", (t2-t1)/N)



"""
Define abstract base classes for solvers to be used in DiffEv and any future solvers.
"""
from abc import ABC, ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import List, Union

from .core.config import Configurable
from .core.h5_support import H5HintedExport
from .data import DataList
from .model import Model

try:
    from numpy.typing import ArrayLike
except ImportError:
    from numpy import ndarray as ArrayLike

class GenxOptimizer(Configurable, H5HintedExport, metaclass=ABCMeta):
    """
    Defines an abstract base class for a optimizer for GenX models.
    DiffEv is implementing this abstraction.
    """
    h5group_name='optimizer'
    _export_ignore = ['opt', 'model']

    @property
    @abstractmethod
    def n_fom_evals(self)->int:
        """Returns teh number of FOM evaluations"""

    @abstractmethod
    def pickle_string(self, clear_evals: bool=False):
        """ Return a pickle string for the object """

    @abstractmethod
    def pickle_load(self, pickled_string: bytes):
        """ Configure object from pickled copy """

    @abstractmethod
    def is_running(self)->bool:
        """Indicate if a refinement is currently running"""

    @abstractmethod
    def get_start_guess(self)->ArrayLike:
        """Return the used start_guess for parameters"""

    @abstractmethod
    def get_model(self)->Model:
        """Return the used model object"""

    @abstractmethod
    def get_fom_log(self)->ArrayLike:
        """Return array of FOM evaluations from fit"""

    @abstractmethod
    def calc_error_bar(self, index:int)->(float, float):
        """Use simple threshold based calculation for errorbar estimation"""

    @abstractmethod
    def project_evals(self, index: int) -> (ArrayLike, ArrayLike):
        """Generate parameter value vs. FOM for previous run of solver"""

    @abstractmethod
    def start_fit(self, model: Model):
        """Start refining the model parameters"""

    @abstractmethod
    def stop_fit(self):
        """Stop refining the model parameters"""

    @abstractmethod
    def resume_fit(self, model: Model):
        """Continue a previously stopped fit"""

    @abstractmethod
    def is_fitted(self)->bool:
        """Indicate if a refinement was done earlier"""

    @abstractmethod
    def is_configured(self)->bool:
        """Has the refinement been setup for a fit"""

    @abstractmethod
    def set_callbacks(self, callbacks: 'GenxOptimizerCallback'):
        """Setup the callback mechanism"""

    @abstractmethod
    def get_callbacks(self)-> 'GenxOptimizerCallback':
        """return the callback mechanism object"""

    @abstractmethod
    def get_result_info(self)->'SolverResultInfo':
        """Return the result info of a previous run"""

    def __repr__(self):
        output=f"{self.__class__.__name__} Optimizer:\n"
        for gname, group in self.opt.groups.items():
            output+='    %s:\n'%gname
            for attr in group:
                if type(attr) is list:
                    output+='        '
                    for ati in attr:
                        try:
                            output += '%s=%s\t'%(ati, getattr(self.opt, ati))
                        except AttributeError:
                            continue
                    output=output[:-1]+'\n'
                else:
                    output+='        %-30s %s\n'%(attr, getattr(self.opt, attr))
        return output

    @property
    def widget(self):
        return self._repr_ipyw_()

    def _ipyw_entry_for_item(self, attr):
        import ipywidgets as ipw
        if type(attr) is list:
            sublist = []
            for subattr in attr:
                try:
                    subentry = self._ipyw_entry_for_item(subattr)
                except AttributeError:
                    continue
                sublist.append(subentry)
            item = ipw.VBox(sublist)
            return item

        val = getattr(self.opt, attr)
        if type(val) is bool:
            item = ipw.Checkbox(value=val, indent=False, description=attr, layout=ipw.Layout(width='24ex'))
            entry = item
        elif type(val) is int:
            entry = ipw.IntText(value=val, layout=ipw.Layout(width='18ex'))
            item = ipw.VBox([ipw.Label(attr), entry])
        elif type(val) is float:
            entry = ipw.FloatText(value=val, layout=ipw.Layout(width='18ex'))
            item = ipw.VBox([ipw.Label(attr), entry])
        elif attr=='method':
            entry = ipw.Dropdown(value=val, options=self.methods, layout=ipw.Layout(width='18ex'))
            item = ipw.VBox([ipw.Label(attr), entry])
        else:
            entry = ipw.Text(value=val, layout=ipw.Layout(width='14ex'))
            item = ipw.VBox([ipw.Label(attr), entry])
        entry.change_item = attr
        entry.observe(self._ipyw_change, names='value')
        return item

    def _repr_ipyw_(self):
        import ipywidgets as ipw
        entries=[]
        for gname, group in self.opt.groups.items():
            gentries=[ipw.HTML("<b>%s:</b>"%gname)]
            for attr in group:
                item=self._ipyw_entry_for_item(attr)
                gentries.append(item)
            entries.append(ipw.VBox(gentries, layout=ipw.Layout(width='26ex')))
        return ipw.VBox([ipw.HTML("<h3>Optimizer Settings:</h3>"), ipw.HBox(entries)])

    def _ipyw_change(self, change):
        setattr(self.opt, change.owner.change_item, change.new)


@dataclass(frozen=True)
class SolverParameterInfo:
    # Used to report model parameter updates
    values: ArrayLike
    new_best: bool
    population: List[ArrayLike]
    max_val: ArrayLike
    min_val: ArrayLike
    fitting: bool

@dataclass(frozen=True)
class SolverUpdateInfo:
    # Used to report model update and result data
    new_best: bool
    fom_value: float
    fom_name: str
    fom_log: ArrayLike
    data: DataList

@dataclass(frozen=True)
class SolverResultInfo(SolverParameterInfo):
    # Used to report the model parameters of the fit result
    start_guess: ArrayLike
    error_message: Union[str, None]

class GenxOptimizerCallback(ABC):
    """
    Defines the callbacks used by the optimizer to give feedback on the state of refinement.
    It defines a set of methods that are used by the solver.
    """

    @abstractmethod
    def text_output(self, text)->None:
        """Send small update string to give status feedback"""

    @abstractmethod
    def plot_output(self, update_data: SolverUpdateInfo)->None:
        """Send full function results back for plotting"""

    @abstractmethod
    def parameter_output(self, param_info: SolverParameterInfo)->None:
        """Update the state of the model parameters"""

    @abstractmethod
    def fitting_ended(self, result_data: SolverResultInfo)->None:
        """Indicate a fit has ended"""

    @abstractmethod
    def autosave(self)->None:
        """Send small update string to give status feedback"""


#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
.. module:: island_lattice.py
   :synopsis: Models to calcualte the reflectivity from a 2D island array

This module provide functions to calculate the reflectivity from gratings using the distorted wave Born approximation
(DWBA) and the kinematical theory.
"""

from .parameters import Int, Float, ComplexArray, Func, Var, List, HasParameters
from .materials import Material
import numpy as np

from . import grating
from genx.core.custom_logging import iprint

q=Var('q', np.array([0, 0, 0]), help="The scattering vector.")
e=Var('wl', 1.54, help="The wavelength of the radiation.")
z=Var('z', 0.0, help="The out-of-plane position in the sample.")
rep=Var('rep', 0, help="The repetition index in a Stack, starting with zero.")

validation_kwargs={'q': np.array([0, 0, 0]), 'z': 0, 'rep': 0}

class Layer(HasParameters):
    validation_kwargs=validation_kwargs
    protected=True

    def __init__(self, **kwargs):
        self.d=Float(0, unit="AA", help="Layer thickness")
        self.sld=ComplexArray(0, unit="1/AA^2", help="The in-plane fourier transform of the scattering length density")
        HasParameters.__init__(self, **kwargs)

class Stack(HasParameters):
    validation_kwargs=validation_kwargs
    protected=True

    def __init__(self, **kwargs):
        try:
            value=kwargs.pop('layers')
        except KeyError:
            value=[]
        self.layers=List(Layer, value=value, help="The layers in the stack")
        self.reps=Int(1, help="Number of repetitions of the stack")
        HasParameters.__init__(self, **kwargs)

    def build_layer_list(self, parameters, kwargs):
        """Builds an array for layer parameters by looping trough the stacks and layers

        Parameters:
            parameters(list of string): name of a valid layer parameters
            kwargs(dict): Keyword argument to be passed to the parameter call
        """
        out=dict()
        [out.__setitem__(par, []) for par in parameters]
        kwargs['rep']=-1
        for n in range(self.reps(**kwargs)):
            kwargs['rep']+=1
            for layer in self.layers:
                # Evaluate the parameter value
                # print kwargs
                [out[par].append(layer.__getattribute__(par)(**kwargs)) for par in parameters]
                kwargs['z']+=layer.d(**kwargs)
        kwargs.pop('rep')
        return out

class Sample(HasParameters):
    validation_kwargs=validation_kwargs

    def __init__(self, **kwargs):
        try:
            value=kwargs.pop('stacks')
        except KeyError:
            value=[]
        self.stacks=List(Stack, value=value, help="The stacks in the sample")
        self.substrate=Layer(d=0.0, sld=0*FLayer())
        self.ambient=Layer(d=0.0, sld=0*FLayer())

        self.a=Float(2*np.pi, unit="AA", help="Unit cell size along the x-direction")
        self.b=Float(2*np.pi, unit="AA", help="Unit cell size along the y-direction")

        # dictionary that holds the last evaluated layer parameters - used for checking the calcs
        self.layer_dic={}

        HasParameters.__init__(self, **kwargs)

    def build_layer_list(self, parameters, kwargs):
        """Builds an array for layer parameters by looping through the stacks and layers

        Parameters:
            parameters(list of string): name of a valid layer parameter
            kwargs(dict): Keyword argument to be passed to the parameter call
        """
        kwargs['z']=0.0
        out=dict()
        [out.__setitem__(par, [self.substrate.__getattribute__(par)(**kwargs)]) for par in parameters]
        for stack in self.stacks:
            stack_dict=stack.build_layer_list(parameters, kwargs)
            [out[par].extend(stack_dict[par]) for par in parameters]
        kwargs['z']=0.0
        [out[par].append(self.ambient.__getattribute__(par)(**kwargs)) for par in parameters]
        return out

    def simulate(self, instrument, h, k, qz):
        self.layer_dic={}
        # Add a q = 0 point for the mean scattering potential
        q=np.c_[h*2*np.pi/self.a(), k*2*np.pi/self.b(), -qz].T
        kwargs={'z': 0.0, 'q': q}
        wavelength=instrument.wl(**kwargs)
        kwargs['wl']=wavelength

        area_uc=(self.a*self.b)(**kwargs)
        dic=self.build_layer_list(['d', 'sld'], kwargs)
        vf, d=4*np.pi*np.array(dic['sld']), np.array(dic['d'])
        kwargs['q']=np.array([0, 0, 0])
        vmean=4*np.pi*np.array(self.build_layer_list(['sld'], kwargs)['sld'])/area_uc
        self.layer_dic['sld_mean']=vmean/4/np.pi
        self.layer_dic['d']=d
        iprint(vmean.shape, vf.shape)
        k_in, k_out=instrument.reflectometer_3axis_kinout(q)
        R=grating.coherent_refl(k_in, k_out, wavelength, vf.T, vmean, d, area_uc, kin=False)
        return R

class Instrument(HasParameters):
    validation_kwargs=validation_kwargs

    def __init__(self, **kwargs):
        self.wl=Float(1.54, unit="AA", help="Wavelength of the used radiation")
        HasParameters.__init__(self, **kwargs)

    def reflectometer_3axis_kinout(self, q):
        """Tranforms q values to k_in and k_out vectors for a reflectometer setup with a sample rotation circle.

        Parameters:
            q(array): an array for qx, qy and qz, size (3xM)

        Returns:
            k_in (array): The incomming wavevector, size (3xM)
            k_out (array): The outgoing wavevector, size (3xM)
        """
        qx=q[0]
        qy=q[1]
        qz=q[2]
        k=2*np.pi/self.wl()
        qp=np.sqrt(qx*qx+qy*qy)
        q_abs=np.sqrt(qx*qx+qy*qy+qz*qz)
        phi=np.arctan2(qy, qx)
        cos_phi=np.cos(phi)
        sin_phi=np.sin(phi)

        apb=np.arcsin(q_abs/2.0/k)
        amb=np.arctan2(qp, -qz)%np.pi
        alpha=apb+amb
        beta=apb-amb
        cos_alpha=np.cos(alpha)
        cos_beta=np.cos(beta)
        k_in=k*np.c_[cos_alpha*cos_phi, cos_alpha*sin_phi, np.sin(alpha)].T
        k_out=k*np.c_[cos_beta*cos_phi, cos_beta*sin_phi, -np.sin(beta)].T
        return k_in, k_out

class FLayer(Func):
    """The in-plane fourier transform of an in-plane infinite layer"""

    def function(self, **kwargs):
        q=kwargs['q']
        return np.where(np.bitwise_and(q[0]==0, q[1]==0), 1.0, 0.0)

class FCircle(Func):
    """The in-plane fourier transform of a circle"""

    def __init__(self, **kwargs):
        self.r=Float(1.0, help="The radius of the circle")
        Func.__init__(self, **kwargs)

    def function(self, **kwargs):
        q=kwargs['q']
        r=self.r(**kwargs)
        q_r=np.sqrt(q[0]**2+q[1]**2)
        return 2*np.pi*r**2*grating.jinc(q_r*r)

if __name__=='__main__':
    radius=Float(1.0)
    slope=Float(1)
    mat_Fe=Material('Fe', density=7.87)
    Fe=Layer(d=rep*0.1+10.0, sld=mat_Fe.sld_x*FCircle(r=radius+z*slope))
    Fe.d=20.0
    q=np.array([0, 0, 0])
    iprint(Fe.sld(q=q))
    iprint(Fe.d)
    ML=Stack(reps=10, layers=[Fe, ])

    s=Sample(stacks=[ML, ])
    iprint(isinstance(s, HasParameters), s.__class__.__name__)
    inst=Instrument()

    iprint('D with a gradient: ', s.build_layer_list(['d'], {})['d'])

    # print s.stacks[0].layers[0]
    qz=np.arange(0, 0.1, 0.001)
    s.simulate(inst, qz*0+0.0, qz*0+0.01, qz)
